# future
from __future__ import annotations

# stdlib
import base64
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from enum import Enum
from getpass import getpass
import json
import logging
from typing import Any
from typing import TYPE_CHECKING
from typing import cast

# third party
from argon2 import PasswordHasher
from cachetools import TTLCache
from cachetools import cached
from pydantic import field_validator
import requests
from requests import Response
from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # type: ignore[import-untyped]
from typing_extensions import Self

# relative
from .. import __version__
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..protocol.data_protocol import DataProtocol
from ..protocol.data_protocol import PROTOCOL_TYPE
from ..protocol.data_protocol import get_data_protocol
from ..serde.deserialize import _deserialize
from ..serde.serializable import serializable
from ..serde.serialize import _serialize
from ..server.credentials import SyftSigningKey
from ..server.credentials import SyftVerifyKey
from ..server.credentials import UserLoginCredentials
from ..service.context import ServerServiceContext
from ..service.metadata.server_metadata import ServerMetadata
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.user.user import UserCreate
from ..service.user.user import UserPrivateKey
from ..service.user.user import UserView
from ..service.user.user_roles import ServiceRole
from ..service.user.user_service import UserService
from ..types.server_url import ServerURL
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.uid import UID
from ..util.telemetry import instrument
from ..util.util import prompt_warning_message
from ..util.util import thread_ident
from ..util.util import verify_tls
from .api import APIModule
from .api import APIRegistry
from .api import SignedSyftAPICall
from .api import SyftAPI
from .api import SyftAPICall
from .api import debox_signed_syftapicall_response
from .connection import ServerConnection
from .protocol import SyftProtocol

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # relative
    from ..service.network.server_peer import ServerPeer


def upgrade_tls(url: ServerURL, response: Response) -> ServerURL:
    try:
        if response.url.startswith("https://") and url.protocol == "http":
            # we got redirected to https
            https_url = ServerURL.from_url(response.url).with_path("")
            logger.debug(f"ServerURL Upgraded to HTTPS. {https_url}")
            return https_url
    except Exception as e:
        print(f"Failed to upgrade to HTTPS. {e}")
    return url


def forward_message_to_proxy(
    make_call: Callable,
    proxy_target_uid: UID,
    path: str,
    credentials: SyftSigningKey | None = None,
    args: list | None = None,
    kwargs: dict | None = None,
) -> Any | SyftError:
    kwargs = {} if kwargs is None else kwargs
    args = [] if args is None else args
    call = SyftAPICall(
        server_uid=proxy_target_uid,
        path=path,
        args=args,
        kwargs=kwargs,
        blocking=True,
    )

    if credentials is None:
        # generate a random signing key
        credentials = SyftSigningKey.generate()

    signed_message: SignedSyftAPICall = call.sign(credentials=credentials)
    signed_result = make_call(signed_message)
    response = debox_signed_syftapicall_response(signed_result)

    return response


API_PATH = "/api/v2"
DEFAULT_SYFT_UI_PORT = 80
DEFAULT_SYFT_UI_ADDRESS = f"http://localhost:{DEFAULT_SYFT_UI_PORT}"
INTERNAL_PROXY_TO_RATHOLE = "http://proxy:80/rtunnel/"


class Routes(Enum):
    ROUTE_METADATA = f"{API_PATH}/metadata"
    ROUTE_API = f"{API_PATH}/api"
    ROUTE_LOGIN = f"{API_PATH}/login"
    ROUTE_REGISTER = f"{API_PATH}/register"
    ROUTE_API_CALL = f"{API_PATH}/api_call"
    ROUTE_BLOB_STORE = "/blob"
    STREAM = f"{API_PATH}/stream"


@serializable(attrs=["proxy_target_uid", "url", "rtunnel_token"])
class HTTPConnection(ServerConnection):
    __canonical_name__ = "HTTPConnection"
    __version__ = SYFT_OBJECT_VERSION_1

    url: ServerURL
    proxy_target_uid: UID | None = None
    routes: type[Routes] = Routes
    session_cache: Session | None = None
    headers: dict[str, str] | None = None
    rtunnel_token: str | None = None

    @field_validator("url", mode="before")
    @classmethod
    def make_url(cls, v: Any) -> Any:
        return (
            ServerURL.from_url(v).as_container_host()
            if isinstance(v, str | ServerURL)
            else v
        )

    def set_headers(self, headers: dict[str, str]) -> None:
        self.headers = headers

    def with_proxy(self, proxy_target_uid: UID) -> Self:
        return HTTPConnection(
            url=self.url,
            proxy_target_uid=proxy_target_uid,
            rtunnel_token=self.rtunnel_token,
        )

    def stream_via(self, proxy_uid: UID, url_path: str) -> ServerURL:
        # Update the presigned url path to
        # <gatewayurl>/<peer_uid>/<presigned_url>
        # url_path_bytes = _serialize(url_path, to_bytes=True)

        url_path_str = base64.urlsafe_b64encode(url_path.encode()).decode()
        stream_url_path = f"{self.routes.STREAM.value}/{proxy_uid}/{url_path_str}/"
        return self.url.with_path(stream_url_path)

    def get_cache_key(self) -> str:
        return str(self.url)

    @property
    def api_url(self) -> ServerURL:
        return self.url.with_path(self.routes.ROUTE_API_CALL.value)

    def to_blob_route(self, path: str, **kwargs: Any) -> ServerURL:
        _path = self.routes.ROUTE_BLOB_STORE.value + path
        return self.url.with_path(_path)

    @property
    def session(self) -> Session:
        if self.session_cache is None:
            session = requests.Session()
            retry = Retry(total=3, backoff_factor=0.5)
            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)
            self.session_cache = session
        return self.session_cache

    def _make_get(
        self, path: str, params: dict | None = None, stream: bool = False
    ) -> bytes | Iterable:
        if params is None:
            return self._make_get_no_params(path, stream=stream)

        url = self.url

        if self.rtunnel_token:
            self.headers = {} if self.headers is None else self.headers
            url = ServerURL.from_url(INTERNAL_PROXY_TO_RATHOLE)
            self.headers["Host"] = self.url.host_or_ip

        url = url.with_path(path)

        response = self.session.get(
            str(url),
            headers=self.headers,
            verify=verify_tls(),
            proxies={},
            params=params,
            stream=stream,
        )
        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch {url}. Response returned with code {response.status_code}"
            )

        # upgrade to tls if available
        self.url = upgrade_tls(self.url, response)

        return response.content

    @cached(cache=TTLCache(maxsize=128, ttl=300))
    def _make_get_no_params(self, path: str, stream: bool = False) -> bytes | Iterable:
        url = self.url

        if self.rtunnel_token:
            self.headers = {} if self.headers is None else self.headers
            url = ServerURL.from_url(INTERNAL_PROXY_TO_RATHOLE)
            self.headers["Host"] = self.url.host_or_ip

        url = url.with_path(path)

        response = self.session.get(
            str(url),
            headers=self.headers,
            verify=verify_tls(),
            proxies={},
            stream=stream,
        )
        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch {url}. Response returned with code {response.status_code}"
            )

        # upgrade to tls if available
        self.url = upgrade_tls(self.url, response)

        if stream:
            return response.iter_content(chunk_size=None)

        return response.content

    def _make_put(
        self, path: str, data: bytes | Generator, stream: bool = False
    ) -> Response:
        url = self.url

        if self.rtunnel_token:
            url = ServerURL.from_url(INTERNAL_PROXY_TO_RATHOLE)
            self.headers = {} if self.headers is None else self.headers
            self.headers["Host"] = self.url.host_or_ip

        url = url.with_path(path)
        response = self.session.put(
            str(url),
            verify=verify_tls(),
            proxies={},
            data=data,
            headers=self.headers,
            stream=stream,
        )
        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch {url}. Response returned with code {response.status_code}"
            )

        # upgrade to tls if available
        self.url = upgrade_tls(self.url, response)

        return response

    def _make_post(
        self,
        path: str,
        json: dict[str, Any] | None = None,
        data: bytes | None = None,
    ) -> bytes:
        url = self.url

        if self.rtunnel_token:
            url = ServerURL.from_url(INTERNAL_PROXY_TO_RATHOLE)
            self.headers = {} if self.headers is None else self.headers
            self.headers["Host"] = self.url.host_or_ip

        url = url.with_path(path)
        response = self.session.post(
            str(url),
            headers=self.headers,
            verify=verify_tls(),
            json=json,
            proxies={},
            data=data,
        )
        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch {url}. Response returned with code {response.status_code}"
            )

        # upgrade to tls if available
        self.url = upgrade_tls(self.url, response)

        return response.content

    def stream_data(self, credentials: SyftSigningKey) -> Response:
        url = self.url.with_path(self.routes.STREAM.value)
        response = self.session.get(
            str(url), verify=verify_tls(), proxies={}, stream=True, headers=self.headers
        )
        return response

    def get_server_metadata(
        self, credentials: SyftSigningKey
    ) -> ServerMetadataJSON | SyftError:
        if self.proxy_target_uid:
            response = forward_message_to_proxy(
                make_call=self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="metadata",
                credentials=credentials,
            )
            return response
        else:
            response = self._make_get(self.routes.ROUTE_METADATA.value)
            metadata_json = json.loads(response)
            return ServerMetadataJSON(**metadata_json)

    def get_api(
        self,
        credentials: SyftSigningKey,
        communication_protocol: int,
        metadata: ServerMetadataJSON | None = None,
    ) -> SyftAPI:
        params = {
            "verify_key": str(credentials.verify_key),
            "communication_protocol": communication_protocol,
        }
        if self.proxy_target_uid:
            obj = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="api",
                kwargs={
                    "credentials": credentials,
                    "communication_protocol": communication_protocol,
                },
                credentials=credentials,
            )
        else:
            content = self._make_get(self.routes.ROUTE_API.value, params=params)
            obj = _deserialize(content, from_bytes=True)
        obj.connection = self
        obj.signing_key = credentials
        obj.communication_protocol = communication_protocol
        obj.metadata = metadata
        if self.proxy_target_uid:
            obj.server_uid = self.proxy_target_uid
        return cast(SyftAPI, obj)

    def login(
        self,
        email: str,
        password: str,
    ) -> SyftSigningKey | None:
        credentials = {"email": email, "password": password}
        if self.proxy_target_uid:
            obj = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="login",
                kwargs=credentials,
            )
        else:
            response = self._make_post(self.routes.ROUTE_LOGIN.value, credentials)
            obj = _deserialize(response, from_bytes=True)

        return obj

    def register(self, new_user: UserCreate) -> SyftSigningKey:
        data = _serialize(new_user, to_bytes=True)
        if self.proxy_target_uid:
            response = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="register",
                kwargs={"new_user": new_user},
            )
        else:
            response = self._make_post(self.routes.ROUTE_REGISTER.value, data=data)
            response = _deserialize(response, from_bytes=True)
        return response

    def make_call(self, signed_call: SignedSyftAPICall) -> Any | SyftError:
        msg_bytes: bytes = _serialize(obj=signed_call, to_bytes=True)

        if self.rtunnel_token:
            api_url = ServerURL.from_url(INTERNAL_PROXY_TO_RATHOLE)
            api_url = api_url.with_path(self.routes.ROUTE_API_CALL.value)
            self.headers = {} if self.headers is None else self.headers
            self.headers["Host"] = self.url.host_or_ip
        else:
            api_url = self.api_url

        response = requests.post(  # nosec
            url=api_url,
            data=msg_bytes,
            headers=self.headers,
        )

        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch metadata. Response returned with code {response.status_code}"
            )

        result = _deserialize(response.content, from_bytes=True)
        return result

    def __repr__(self) -> str:
        return f"{type(self).__name__}: {self.url}"

    def __str__(self) -> str:
        return f"{type(self).__name__}: {self.url}"

    def __hash__(self) -> int:
        return hash(self.proxy_target_uid) + hash(self.url)

    def get_client_type(self) -> type[SyftClient] | SyftError:
        # TODO: Rasswanth, should remove passing in credentials
        # when metadata are proxy forwarded in the server routes
        # in the gateway fixes PR
        # relative
        from .datasite_client import DatasiteClient
        from .enclave_client import EnclaveClient
        from .gateway_client import GatewayClient

        metadata = self.get_server_metadata(credentials=SyftSigningKey.generate())
        if metadata.server_type == ServerType.DATASITE.value:
            return DatasiteClient
        elif metadata.server_type == ServerType.GATEWAY.value:
            return GatewayClient
        elif metadata.server_type == ServerType.ENCLAVE.value:
            return EnclaveClient
        else:
            return SyftError(message=f"Unknown server type {metadata.server_type}")


@serializable()
class PythonConnection(ServerConnection):
    __canonical_name__ = "PythonConnection"
    __version__ = SYFT_OBJECT_VERSION_1

    server: AbstractServer
    proxy_target_uid: UID | None = None

    def with_proxy(self, proxy_target_uid: UID) -> Self:
        return PythonConnection(server=self.server, proxy_target_uid=proxy_target_uid)

    def get_server_metadata(
        self, credentials: SyftSigningKey
    ) -> ServerMetadataJSON | SyftError:
        if self.proxy_target_uid:
            response = forward_message_to_proxy(
                make_call=self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="metadata",
                credentials=credentials,
            )
            return response
        else:
            return self.server.metadata.to(ServerMetadataJSON)

    def to_blob_route(self, path: str, host: str | None = None) -> ServerURL:
        # TODO: FIX!
        if host is not None:
            return ServerURL(host_or_ip=host, port=8333).with_path(path)
        else:
            return ServerURL(port=8333).with_path(path)

    def get_api(
        self,
        credentials: SyftSigningKey,
        communication_protocol: int,
        metadata: ServerMetadataJSON | None = None,
    ) -> SyftAPI:
        # todo: its a bit odd to identify a user by its verify key maybe?
        if self.proxy_target_uid:
            obj = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="api",
                kwargs={
                    "credentials": credentials,
                    "communication_protocol": communication_protocol,
                },
                credentials=credentials,
            )
        else:
            obj = self.server.get_api(
                for_user=credentials.verify_key,
                communication_protocol=communication_protocol,
            )
        obj.connection = self
        obj.signing_key = credentials
        obj.communication_protocol = communication_protocol
        obj.metadata = metadata
        if self.proxy_target_uid:
            obj.server_uid = self.proxy_target_uid
        return obj

    def get_cache_key(self) -> str:
        return str(self.server.id)

    def exchange_credentials(self, email: str, password: str) -> UserPrivateKey | None:
        context = self.server.get_unauthed_context(
            login_credentials=UserLoginCredentials(email=email, password=password)
        )
        method = self.server.get_method_with_context(
            UserService.exchange_credentials, context
        )
        result = method()
        return result

    def login(
        self,
        email: str,
        password: str,
    ) -> SyftSigningKey | None:
        if self.proxy_target_uid:
            obj = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="login",
                kwargs={"email": email, "password": password},
            )

        else:
            obj = self.exchange_credentials(email=email, password=password)
        return obj

    def register(self, new_user: UserCreate) -> SyftSigningKey | None:
        if self.proxy_target_uid:
            response = forward_message_to_proxy(
                self.make_call,
                proxy_target_uid=self.proxy_target_uid,
                path="register",
                kwargs={"new_user": new_user},
            )
        else:
            service_context = ServerServiceContext(server=self.server)
            method = self.server.get_service_method(UserService.register)
            response = method(context=service_context, new_user=new_user)
        return response

    def make_call(self, signed_call: SignedSyftAPICall) -> Any | SyftError:
        return self.server.handle_api_call(signed_call)

    def __repr__(self) -> str:
        return f"{type(self).__name__}"

    def __str__(self) -> str:
        return f"{type(self).__name__}"

    def get_client_type(self) -> type[SyftClient] | SyftError:
        # relative
        from .datasite_client import DatasiteClient
        from .enclave_client import EnclaveClient
        from .gateway_client import GatewayClient

        metadata = self.get_server_metadata(credentials=SyftSigningKey.generate())
        if metadata.server_type == ServerType.DATASITE.value:
            return DatasiteClient
        elif metadata.server_type == ServerType.GATEWAY.value:
            return GatewayClient
        elif metadata.server_type == ServerType.ENCLAVE.value:
            return EnclaveClient
        else:
            return SyftError(message=f"Unknown server type {metadata.server_type}")


@instrument
@serializable(canonical_name="SyftClient", version=1)
class SyftClient:
    connection: ServerConnection
    metadata: ServerMetadataJSON | None
    credentials: SyftSigningKey | None
    __logged_in_user: str = ""
    __logged_in_username: str = ""
    __user_role: ServiceRole = ServiceRole.NONE

    # informs getattr does not have nasty side effects
    __syft_allow_autocomplete__ = [
        "api",
        "code",
        "jobs",
        "users",
        "settings",
        "notifications",
        "custom_api",
    ]

    def __init__(
        self,
        connection: ServerConnection,
        metadata: ServerMetadataJSON | None = None,
        credentials: SyftSigningKey | None = None,
        api: SyftAPI | None = None,
    ) -> None:
        self.connection = connection
        self.metadata = metadata
        self.credentials: SyftSigningKey | None = credentials
        self._api = api
        self.services: APIModule | None = None
        self.communication_protocol: int | str | None = None
        self.current_protocol: int | str | None = None

        self.post_init()

    def get_env(self) -> str:
        return self.api.services.metadata.get_env()

    def post_init(self) -> None:
        if self.metadata is None:
            self._fetch_server_metadata(self.credentials)
        self.metadata = cast(ServerMetadataJSON, self.metadata)
        self.communication_protocol = self._get_communication_protocol(
            self.metadata.supported_protocols
        )

    def set_headers(self, headers: dict[str, str]) -> None | SyftError:
        if isinstance(self.connection, HTTPConnection):
            self.connection.set_headers(headers)
            return None
        return SyftError(  # type: ignore
            message="Incompatible connection type."
            + f"Expected HTTPConnection, got {type(self.connection)}"
        )

    def _get_communication_protocol(
        self, protocols_supported_by_server: list
    ) -> int | str:
        data_protocol: DataProtocol = get_data_protocol()
        protocols_supported_by_client: list[PROTOCOL_TYPE] = (
            data_protocol.supported_protocols
        )

        self.current_protocol = data_protocol.latest_version
        common_protocols = set(protocols_supported_by_client).intersection(
            protocols_supported_by_server
        )

        if len(common_protocols) == 0:
            raise Exception(
                "No common communication protocol found between the client and the server."
            )

        if "dev" in common_protocols:
            return "dev"
        return max(common_protocols)

    def create_project(
        self, name: str, description: str, user_email_address: str
    ) -> Any:
        # relative
        from ..service.project.project import ProjectSubmit

        project_create = ProjectSubmit(
            name=name,
            description=description,
            shareholders=[self],
            user_email_address=user_email_address,
            members=[self],
        )
        project = project_create.send()
        return project

    @property
    def authed(self) -> bool:
        return bool(self.credentials)

    @property
    def logged_in_user(self) -> str | None:
        return self.__logged_in_user

    @property
    def logged_in_username(self) -> str | None:
        return self.__logged_in_username

    @property
    def user_role(self) -> ServiceRole:
        return self.__user_role

    @property
    def verify_key(self) -> SyftVerifyKey:
        if self.credentials is None:
            raise ValueError("SigningKey not set on client")
        return self.credentials.verify_key

    @classmethod
    def from_url(cls, url: str | ServerURL) -> Self:
        return cls(connection=HTTPConnection(url=ServerURL.from_url(url)))

    @classmethod
    def from_server(cls, server: AbstractServer) -> Self:
        return cls(connection=PythonConnection(server=server))

    @property
    def name(self) -> str | None:
        return self.metadata.name if self.metadata else None

    @property
    def id(self) -> UID | None:
        return UID.from_string(self.metadata.id) if self.metadata else None

    @property
    def icon(self) -> str:
        return "📡"

    @property
    def peer(self) -> Any:
        # relative
        from ..service.network.network_service import ServerPeer

        return ServerPeer.from_client(self)

    @property
    def route(self) -> Any:
        return self.connection.route

    @property
    def api(self) -> SyftAPI:
        # invalidate API
        if self._api is None or (self._api.signing_key != self.credentials):
            self._fetch_api(self.credentials)
        return cast(SyftAPI, self._api)  # we are sure self._api is not None after fetch

    def guest(self) -> Self:
        return self.__class__(
            connection=self.connection,
            credentials=SyftSigningKey.generate(),
            metadata=self.metadata,
        )

    def exchange_route(
        self,
        client: Self,
        protocol: SyftProtocol = SyftProtocol.HTTP,
        reverse_tunnel: bool = False,
    ) -> SyftSuccess | SyftError:
        # relative
        from ..service.network.routes import connection_to_route

        if protocol == SyftProtocol.HTTP:
            self_server_route = connection_to_route(self.connection)
            remote_server_route = connection_to_route(client.connection)
            if client.metadata is None:
                return SyftError(f"client {client}'s metadata is None!")

            return self.api.services.network.exchange_credentials_with(
                self_server_route=self_server_route,
                remote_server_route=remote_server_route,
                remote_server_verify_key=client.metadata.to(ServerMetadata).verify_key,
                reverse_tunnel=reverse_tunnel,
            )
        else:
            raise ValueError(
                f"Invalid Route Exchange SyftProtocol: {protocol}.Supported protocols are {SyftProtocol.all()}"
            )

    @property
    def jobs(self) -> APIModule | None:
        if self.api.has_service("job"):
            return self.api.services.job
        return None

    @property
    def users(self) -> APIModule | None:
        if self.api.has_service("user"):
            return self.api.services.user
        return None

    @property
    def custom_api(self) -> APIModule | None:
        if self.api.has_service("api"):
            return self.api.services.api
        return None

    @property
    def numpy(self) -> APIModule | None:
        if self.api.has_lib("numpy"):
            return self.api.lib.numpy
        return None

    @property
    def settings(self) -> APIModule | None:
        if self.api.has_service("settings"):
            return self.api.services.settings
        return None

    @property
    def notifications(self) -> APIModule | None:
        if self.api.has_service("notifications"):
            return self.api.services.notifications
        return None

    @property
    def notifier(self) -> APIModule | None:
        if self.api.has_service("notifier"):
            return self.api.services.notifier
        return None

    @property
    def peers(self) -> list[ServerPeer] | SyftError | None:
        if self.api.has_service("network"):
            return self.api.services.network.get_all_peers()
        return None

    @property
    def account(self) -> UserView | SyftError | None:
        if self.api.has_service("user"):
            return self.api.services.user.get_current_user()
        return None

    def login_as_guest(self) -> Self:
        _guest_client = self.guest()

        if self.metadata is not None:
            print(
                f"Logged into <{self.name}: {self.metadata.server_side_type.capitalize()}-side "
                f"{self.metadata.server_type.capitalize()}> as GUEST"
            )

        return _guest_client

    def login_as(self, email: str) -> Self:
        user_private_key = self.api.services.user.key_for_email(email=email)
        if not isinstance(user_private_key, UserPrivateKey):
            return user_private_key
        if self.metadata is not None:
            print(
                f"Logged into <{self.name}: {self.metadata.server_side_type.capitalize()}-side "
                f"{self.metadata.server_type.capitalize()}> as {email}"
            )

        return self.__class__(
            connection=self.connection,
            credentials=user_private_key.signing_key,
            metadata=self.metadata,
        )

    def login(
        self,
        email: str | None = None,
        password: str | None = None,
        cache: bool = True,
        register: bool = False,
        **kwargs: Any,
    ) -> Self:
        if email is None:
            email = input("Email: ")
        if password is None:
            password = getpass("Password: ")

        if register:
            self.register(
                email=email, password=password, password_verify=password, **kwargs
            )

        user_private_key = self.connection.login(email=email, password=password)
        if isinstance(user_private_key, SyftError):
            return user_private_key

        signing_key = None if user_private_key is None else user_private_key.signing_key

        client = self.__class__(
            connection=self.connection,
            metadata=self.metadata,
            credentials=signing_key,
        )

        client.__logged_in_user = email

        if user_private_key is not None and client.users is not None:
            client.__user_role = user_private_key.role
            client.__logged_in_username = client.users.get_current_user().name

        if signing_key is not None and client.metadata is not None:
            print(
                f"Logged into <{client.name}: {client.metadata.server_side_type.capitalize()} side "
                f"{client.metadata.server_type.capitalize()}> as <{email}>"
            )
            # relative
            from ..server.server import get_default_root_password

            if password == get_default_root_password():
                message = (
                    "You are using a default password. Please change the password "
                    "using `[your_client].account.set_password([new_password])`."
                )
                prompt_warning_message(message)

            if cache:
                SyftClientSessionCache.add_client(
                    email=email,
                    password=password,
                    connection=client.connection,
                    syft_client=client,
                )
                # Adding another cache storage
                # as this would be useful in retrieving unique clients
                # server uid and verify key are not individually unique
                # both the combination of server uid and verify key are unique
                # which could be used to identity a client uniquely of any given server
                # TODO: It would be better to have a single cache storage
                # combining both email, password and verify key and uid
                SyftClientSessionCache.add_client_by_uid_and_verify_key(
                    verify_key=signing_key.verify_key,
                    server_uid=client.id,
                    syft_client=client,
                )

        # relative
        from ..server.server import CODE_RELOADER

        thread_id = thread_ident()
        if thread_id is not None:
            CODE_RELOADER[thread_id] = client._reload_user_code

        return client

    def _reload_user_code(self) -> None:
        # relative
        from ..service.code.user_code import load_approved_policy_code

        user_code_items = self.code.get_all_for_user()
        load_approved_policy_code(user_code_items=user_code_items, context=None)

    def register(
        self,
        name: str,
        email: str | None = None,
        password: str | None = None,
        password_verify: str | None = None,
        institution: str | None = None,
        website: str | None = None,
    ) -> SyftError | SyftSigningKey | None:
        if not email:
            email = input("Email: ")
        if not password:
            password = getpass("Password: ")
        if not password_verify:
            password_verify = getpass("Confirm Password: ")
        if password != password_verify:
            return SyftError(message="Passwords do not match")

        try:
            new_user = UserCreate(
                name=name,
                email=email,
                password=password,
                password_verify=password_verify,
                institution=institution,
                website=website,
                created_by=(
                    None if self.__user_role == ServiceRole.GUEST else self.credentials
                ),
            )
        except Exception as e:
            return SyftError(message=str(e))

        if (
            self.metadata
            and self.metadata.server_side_type == ServerSideType.HIGH_SIDE.value
        ):
            message = (
                "You're registering a user to a high side "
                f"{self.metadata.server_type}, which could "
                "host datasets with private information."
            )
            if self.metadata.show_warnings and not prompt_warning_message(
                message=message
            ):
                return None

        response = self.connection.register(new_user=new_user)
        if isinstance(response, tuple):
            response = response[0]
        return response

    def __hash__(self) -> int:
        return hash(self.id) + hash(self.connection)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SyftClient):
            return False
        return (
            self.metadata == other.metadata
            and self.connection == other.connection
            and self.credentials == other.credentials
        )

    def __repr__(self) -> str:
        proxy_target_uid = None
        if self.connection and self.connection.proxy_target_uid:
            proxy_target_uid = self.connection.proxy_target_uid
        client_type = type(self).__name__
        uid = self.id
        if proxy_target_uid:
            client_type = "ProxyClient"
            uid = proxy_target_uid
            return f"<{client_type} - <{uid}>: via {self.connection}>"
        return f"<{client_type} - {self.name} <{uid}>: {self.connection}>"

    def _fetch_server_metadata(self, credentials: SyftSigningKey) -> None:
        metadata = self.connection.get_server_metadata(credentials=credentials)
        if isinstance(metadata, ServerMetadataJSON):
            metadata.check_version(__version__)
            self.metadata = metadata

    def _fetch_api(self, credentials: SyftSigningKey) -> SyftAPI:
        _api: SyftAPI = self.connection.get_api(
            credentials=credentials,
            communication_protocol=self.communication_protocol,
            metadata=self.metadata,
        )
        self._fetch_server_metadata(self.credentials)

        def refresh_callback() -> SyftAPI:
            return self._fetch_api(self.credentials)

        _api.refresh_api_callback = refresh_callback

        if self.credentials is None:
            raise ValueError(f"{self}'s credentials (signing key) is None!")

        APIRegistry.set_api_for(
            server_uid=self.id,
            user_verify_key=self.credentials.verify_key,
            api=_api,
        )
        self._api = _api
        self._api.metadata = self.metadata
        self.services = _api.services
        return _api


@instrument
def connect(
    url: str | ServerURL = DEFAULT_SYFT_UI_ADDRESS,
    server: AbstractServer | None = None,
    port: int | None = None,
) -> SyftClient:
    if server:
        connection = PythonConnection(server=server)
    else:
        url = ServerURL.from_url(url)
        if isinstance(port, int | str):
            url.set_port(int(port))
        connection = HTTPConnection(url=url)

    client_type = connection.get_client_type()

    if isinstance(client_type, SyftError):
        return client_type

    return client_type(connection=connection)


@instrument
def register(
    url: str | ServerURL,
    port: int,
    name: str,
    email: str,
    password: str,
    institution: str | None = None,
    website: str | None = None,
) -> SyftError | SyftSigningKey | None:
    guest_client = connect(url=url, port=port)
    return guest_client.register(
        name=name,
        email=email,
        password=password,
        institution=institution,
        website=website,
    )


@instrument
def login_as_guest(
    # HTTPConnection
    url: str | ServerURL = DEFAULT_SYFT_UI_ADDRESS,
    port: int | None = None,
    # PythonConnection
    server: AbstractServer | None = None,
    verbose: bool = True,
) -> SyftClient:
    _client = connect(
        url=url,
        server=server,
        port=port,
    )

    if isinstance(_client, SyftError):
        return _client

    if verbose and _client.metadata is not None:
        print(
            f"Logged into <{_client.name}: {_client.metadata.server_side_type.capitalize()}-"
            f"side {_client.metadata.server_type.capitalize()}> as GUEST"
        )

    return _client.guest()


@instrument
def login(
    email: str,
    # HTTPConnection
    url: str | ServerURL = DEFAULT_SYFT_UI_ADDRESS,
    port: int | None = None,
    # PythonConnection
    server: AbstractServer | None = None,
    password: str | None = None,
    cache: bool = True,
) -> SyftClient:
    _client = connect(
        url=url,
        server=server,
        port=port,
    )

    if isinstance(_client, SyftError):
        return _client

    connection = _client.connection

    login_credentials = None

    if not password:
        password = getpass("Password: ")
    login_credentials = UserLoginCredentials(email=email, password=password)

    if cache and login_credentials:
        _client_cache = SyftClientSessionCache.get_client(
            login_credentials.email,
            login_credentials.password,
            connection=connection,
        )
        if _client_cache:
            print(
                f"Using cached client for {_client.name} as <{login_credentials.email}>"
            )
            _client = _client_cache

    if not _client.authed and login_credentials:
        _client = _client.login(
            email=login_credentials.email,
            password=login_credentials.password,
            cache=cache,
        )

    return _client


class SyftClientSessionCache:
    __credentials_store__: dict = {}
    __cache_key_format__ = "{email}-{password}-{connection}"
    __client_cache__: dict = {}

    @classmethod
    def _get_key(cls, email: str, password: str, connection: str) -> str:
        key = cls.__cache_key_format__.format(
            email=email, password=password, connection=connection
        )
        ph = PasswordHasher()
        return ph.hash(key)

    @classmethod
    def add_client(
        cls,
        email: str,
        password: str,
        connection: ServerConnection,
        syft_client: SyftClient,
    ) -> None:
        hash_key = cls._get_key(email, password, connection.get_cache_key())
        cls.__credentials_store__[hash_key] = syft_client
        cls.__client_cache__[syft_client.id] = syft_client

    @classmethod
    def add_client_by_uid_and_verify_key(
        cls,
        verify_key: SyftVerifyKey,
        server_uid: UID,
        syft_client: SyftClient,
    ) -> None:
        hash_key = str(server_uid) + str(verify_key)
        cls.__client_cache__[hash_key] = syft_client

    @classmethod
    def get_client_by_uid_and_verify_key(
        cls, verify_key: SyftVerifyKey, server_uid: UID
    ) -> SyftClient | None:
        hash_key = str(server_uid) + str(verify_key)
        return cls.__client_cache__.get(hash_key, None)

    @classmethod
    def get_client(
        cls, email: str, password: str, connection: ServerConnection
    ) -> SyftClient | None:
        # we have some bugs here so lets disable until they are fixed.
        return None
        # hash_key = cls._get_key(email, password, connection.get_cache_key())
        # return cls.__credentials_store__.get(hash_key, None)

    @classmethod
    def get_client_for_server_uid(cls, server_uid: UID) -> SyftClient | None:
        return cls.__client_cache__.get(server_uid, None)



# stdlib
from collections.abc import Callable
from typing import Any

# third party
from pydantic import EmailStr
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..server.credentials import SyftVerifyKey
from ..service.context import AuthedServiceContext
from ..service.context import ServerServiceContext
from .server_url import ServerURL
from .syft_object import Context
from .syft_object import SyftBaseObject
from .syft_object_registry import SyftObjectRegistry
from .uid import UID


class NotNone:
    pass


class TransformContext(Context):
    output: dict[str, Any] | None = None
    server: AbstractServer | None = None
    credentials: SyftVerifyKey | None = None
    obj: Any | None = None

    @classmethod
    def from_context(cls, obj: Any, context: Context | None = None) -> Self:
        t_context = cls()
        t_context.obj = obj
        try:
            t_context.output = obj.to_dict(exclude_empty=True)
        except Exception:
            t_context.output = dict(obj)
        if context is None:
            return t_context
        if hasattr(context, "credentials"):
            t_context.credentials = context.credentials
        if hasattr(context, "server"):
            t_context.server = context.server
        return t_context

    def to_server_context(self) -> ServerServiceContext:
        if self.credentials:
            return AuthedServiceContext(
                server=self.server, credentials=self.credentials
            )
        if self.server:
            return ServerServiceContext(server=self.server)
        return Context()


def geteitherattr(
    _self: Any, output: dict, key: str, default: Any = NotNone
) -> Any | None:
    if key in output:
        return output[key]
    if default == NotNone:
        return getattr(_self, key)
    return getattr(_self, key, default)


def make_set_default(key: str, value: Any) -> Callable:
    def set_default(context: TransformContext) -> TransformContext:
        if context.output and not geteitherattr(context.obj, context.output, key, None):
            context.output[key] = value
        return context

    return set_default


def drop(list_keys: list[str]) -> Callable:
    def drop_keys(context: TransformContext) -> TransformContext:
        if context.output:
            for key in list_keys:
                if key in context.output:
                    del context.output[key]
        return context

    return drop_keys


def rename(old_key: str, new_key: str) -> Callable:
    def drop_keys(context: TransformContext) -> TransformContext:
        if context.output:
            context.output[new_key] = geteitherattr(
                context.obj, context.output, old_key
            )
            if old_key in context.output:
                del context.output[old_key]
        return context

    return drop_keys


def keep(list_keys: list[str]) -> Callable:
    def drop_keys(context: TransformContext) -> TransformContext:
        if context.output is None:
            return context

        for key in list_keys:
            if key not in context.output:
                context.output[key] = getattr(context.obj, key, None)

        keys = list(context.output.keys())

        for key in keys:
            if key not in list_keys and key in context.output:
                del context.output[key]

        return context

    return drop_keys


def convert_types(
    list_keys: list[str], types: type | list[type]
) -> Callable[[TransformContext], TransformContext]:
    if not isinstance(types, list):
        types = [types] * len(list_keys)

    if isinstance(types, list) and len(types) != len(list_keys):
        raise Exception("convert types lists must be the same length")

    def run_convert_types(context: TransformContext) -> TransformContext:
        if context.output:
            for key, _type in zip(list_keys, types):
                context.output[key] = _type(
                    geteitherattr(context.obj, context.output, key)
                )
        return context

    return run_convert_types


def generate_id(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context
    if "id" not in context.output or not isinstance(context.output["id"], UID):
        context.output["id"] = UID()
    return context


def generate_action_object_id(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context
    if "action_object_id" not in context.output or not isinstance(
        context.output["action_object_id"], UID
    ):
        context.output["action_object_id"] = UID()
    return context


def validate_url(context: TransformContext) -> TransformContext:
    if context.output and context.output["url"] is not None:
        context.output["url"] = ServerURL.from_url(context.output["url"]).url_no_port
    return context


def validate_email(context: TransformContext) -> TransformContext:
    if context.output and context.output["email"] is not None:
        EmailStr._validate(context.output["email"])
    return context


def str_url_to_server_url(context: TransformContext) -> TransformContext:
    if context.output:
        url = context.output.get("url", None)
        if url is not None and isinstance(url, str):
            context.output["url"] = ServerURL.from_url(str)
    return context


def add_credentials_for_key(key: str) -> Callable:
    def add_credentials(context: TransformContext) -> TransformContext:
        if context.output is not None:
            context.output[key] = context.credentials
        return context

    return add_credentials


def add_server_uid_for_key(key: str) -> Callable:
    def add_server_uid(context: TransformContext) -> TransformContext:
        if context.output is not None and context.server is not None:
            context.output[key] = context.server.id
        return context

    return add_server_uid


def generate_transform_wrapper(
    klass_from: type, klass_to: type, transforms: list[Callable]
) -> Callable:
    def wrapper(
        self: klass_from,
        context: TransformContext | ServerServiceContext | None = None,
    ) -> klass_to:
        t_context = TransformContext.from_context(obj=self, context=context)
        for transform in transforms:
            t_context = transform(t_context)
        return klass_to(**t_context.output)

    return wrapper


def validate_klass_and_version(
    klass_from: type | str,
    klass_to: type | str,
    version_from: int | None = None,
    version_to: int | None = None,
) -> tuple[str, int | None, str, int | None]:
    if not isinstance(klass_from, type | str):
        raise NotImplementedError(
            "Arguments to `klass_from` should be either of `Type` or `str` type."
        )

    if isinstance(klass_from, str):
        klass_from_str = klass_from
    elif issubclass(klass_from, SyftBaseObject):
        klass_from_str = klass_from.__canonical_name__
        version_from = klass_from.__version__
    else:
        klass_from_str = klass_from.__name__
        version_from = None

    if not isinstance(klass_to, type | str):
        raise NotImplementedError(
            "Arguments to `klass_to` should be either of `Type` or `str` type."
        )

    if isinstance(klass_to, str):
        klass_to_str = klass_to
    elif issubclass(klass_to, SyftBaseObject):
        klass_to_str = klass_to.__canonical_name__
        version_to = klass_to.__version__
    else:
        klass_to_str = klass_to.__name__
        version_to = None

    return klass_from_str, version_from, klass_to_str, version_to


def transform_method(
    klass_from: type | str,
    klass_to: type | str,
    version_from: int | None = None,
    version_to: int | None = None,
) -> Callable:
    (
        klass_from_str,
        version_from,
        klass_to_str,
        version_to,
    ) = validate_klass_and_version(
        klass_from=klass_from,
        version_from=version_from,
        klass_to=klass_to,
        version_to=version_to,
    )

    def decorator(function: Callable) -> Callable:
        SyftObjectRegistry.add_transform(
            klass_from=klass_from_str,
            version_from=version_from,
            klass_to=klass_to_str,
            version_to=version_to,
            method=function,
        )

        return function

    return decorator


def transform(
    klass_from: type | str,
    klass_to: type | str,
    version_from: int | None = None,
    version_to: int | None = None,
) -> Callable:
    (
        klass_from_str,
        version_from,
        klass_to_str,
        version_to,
    ) = validate_klass_and_version(
        klass_from=klass_from,
        version_from=version_from,
        klass_to=klass_to,
        version_to=version_to,
    )

    def decorator(function: Callable) -> Callable:
        transforms = function()

        wrapper = generate_transform_wrapper(
            klass_from=klass_from, klass_to=klass_to, transforms=transforms
        )

        SyftObjectRegistry.add_transform(
            klass_from=klass_from_str,
            version_from=version_from,
            klass_to=klass_to_str,
            version_to=version_to,
            method=wrapper,
        )

        return function

    return decorator



# stdlib
from typing import Any


def as_markdown_code(str: str, lang: str = "python") -> str:
    return f"```{lang}\n{str}\n```"


def as_markdown_python_code(str: str) -> str:
    return as_markdown_code(str, lang="python")


def markdown_as_class_with_fields(
    obj: Any, fields: dict[str, str], set_defaults: bool = True
) -> str:
    if set_defaults:
        # add default properties to start of the dict
        fields = {**{"id": obj.id}, **fields}
    _repr_str = f"class {obj.__class__.__name__}:\n  "
    _repr_str += "\n  ".join([f"{k}: {v}" for k, v in fields.items()])
    return as_markdown_python_code(_repr_str)


class CodeMarkdown:
    def __init__(self, code: str, lang: str = "python"):
        self._code = code
        self._lang = lang

    def _repr_markdown_(self) -> str:
        return as_markdown_code(self._code, self._lang)



# relative
from ..abstract_server import ServerType
from ..serde.serializable import serializable
from .server import Server


@serializable(canonical_name="Gateway", version=1)
class Gateway(Server):
    def post_init(self) -> None:
        self.server_type = ServerType.GATEWAY
        super().post_init()



# stdlib
from typing import Any
from typing import TypeVar
from typing import cast

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.cache_object import CachedSyftObject
from ...types.syft_metaclass import Empty
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_service import ActionService
from ..context import AuthedServiceContext
from ..output.output_service import ExecutionOutput
from ..policy.policy import OutputPolicy
from ..request.request import Request
from ..request.request import SubmitRequest
from ..request.request import SyncedUserCodeStatusChange
from ..request.request import UserCodeStatusChange
from ..request.request_service import RequestService
from ..response import SyftError
from ..response import SyftNotReady
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .user_code import SubmitUserCode
from .user_code import UserCode
from .user_code import UserCodeStatus
from .user_code import UserCodeUpdate
from .user_code import get_code_hash
from .user_code import load_approved_policy_code
from .user_code_stash import UserCodeStash


@instrument
@serializable(canonical_name="UserCodeService", version=1)
class UserCodeService(AbstractService):
    store: DocumentStore
    stash: UserCodeStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = UserCodeStash(store=store)

    @service_method(path="code.submit", name="submit", roles=GUEST_ROLE_LEVEL)
    def submit(
        self, context: AuthedServiceContext, code: SubmitUserCode
    ) -> UserCode | SyftError:
        """Add User Code"""
        result = self._submit(context, code, exists_ok=False)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="User Code Submitted", require_api_update=True)

    def _submit(
        self,
        context: AuthedServiceContext,
        submit_code: SubmitUserCode,
        exists_ok: bool = False,
    ) -> Result[UserCode, str]:
        """
        Submit a UserCode.

        If exists_ok is True, the function will return the existing code if it exists.

        Args:
            context (AuthedServiceContext): context
            submit_code (SubmitUserCode): UserCode to submit
            exists_ok (bool, optional): If True, return the existing code if it exists.
                If false, existing codes returns Err. Defaults to False.

        Returns:
            Result[UserCode, str]: New UserCode or error
        """
        existing_code_or_err = self.stash.get_by_code_hash(
            context.credentials,
            code_hash=get_code_hash(submit_code.code, context.credentials),
        )

        if existing_code_or_err.is_err():
            return existing_code_or_err
        existing_code = existing_code_or_err.ok()
        if existing_code is not None:
            if not exists_ok:
                return Err("The code to be submitted already exists")
            return Ok(existing_code)

        code = submit_code.to(UserCode, context=context)

        result = self._post_user_code_transform_ops(context, code)
        if result.is_err():
            # if the validation fails, we should remove the user code status
            # and code version to prevent dangling status
            root_context = AuthedServiceContext(
                credentials=context.server.verify_key, server=context.server
            )

            if code.status_link is not None:
                _ = context.server.get_service("usercodestatusservice").remove(
                    root_context, code.status_link.object_uid
                )
            return result

        result = self.stash.set(context.credentials, code)
        return result

    @service_method(
        path="code.update",
        name="update",
        roles=ADMIN_ROLE_LEVEL,
        autosplat=["code_update"],
    )
    def update(
        self,
        context: AuthedServiceContext,
        code_update: UserCodeUpdate,
    ) -> SyftSuccess | SyftError:
        code = self.stash.get_by_uid(context.credentials, code_update.id)
        if code.is_err():
            return SyftError(message=code.err())
        code = code.ok()

        result = self.stash.update(context.credentials, code)
        if result.is_err():
            return SyftError(message=str(result.err()))

        if code_update.l0_deny_reason is not Empty:  # type: ignore[comparison-overlap]
            code.l0_deny_reason = code_update.l0_deny_reason

        result = self.stash.update(context.credentials, code)

        if result.is_ok():
            return result.ok()
        return SyftError(message=str(result.err()))

    @service_method(path="code.delete", name="delete", roles=ADMIN_ROLE_LEVEL)
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        """Delete User Code"""
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="User Code Deleted")

    @service_method(
        path="code.get_by_service_func_name",
        name="get_by_service_func_name",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_service_name(
        self, context: AuthedServiceContext, service_func_name: str
    ) -> list[UserCode] | SyftError:
        result = self.stash.get_by_service_func_name(
            context.credentials, service_func_name=service_func_name
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    def _post_user_code_transform_ops(
        self,
        context: AuthedServiceContext,
        user_code: UserCode,
    ) -> Result[UserCode, str]:
        if user_code.output_readers is None:
            return Err(f"there is no verified output readers for {user_code}")
        if user_code.input_owner_verify_keys is None:
            return Err(message=f"there is no verified input owners for {user_code}")
        if not all(
            x in user_code.input_owner_verify_keys for x in user_code.output_readers
        ):
            return Err("outputs can only be distributed to input owners")

        worker_pool_service = context.server.get_service("SyftWorkerPoolService")
        pool_result = worker_pool_service._get_worker_pool(
            context,
            pool_name=user_code.worker_pool_name,
        )

        if isinstance(pool_result, SyftError):
            return Err(pool_result.message)

        # Create a code history
        code_history_service = context.server.get_service("codehistoryservice")
        result = code_history_service.submit_version(context=context, code=user_code)
        if isinstance(result, SyftError):
            return Err(result.message)

        return Ok(user_code)

    def _request_code_execution(
        self,
        context: AuthedServiceContext,
        user_code: UserCode,
        reason: str | None = "",
    ) -> Request | SyftError:
        # Cannot make multiple requests for the same code
        get_by_usercode_id = context.server.get_service_method(
            RequestService.get_by_usercode_id
        )
        existing_requests = get_by_usercode_id(context, user_code.id)
        if isinstance(existing_requests, SyftError):
            return existing_requests
        if len(existing_requests) > 0:
            return SyftError(
                message=f"Request {existing_requests[0].id} already exists for this UserCode. "
                f"Please use the existing request, or submit a new UserCode to create a new request."
            )

        # Users that have access to the output also have access to the code item
        if user_code.output_readers is not None:
            self.stash.add_permissions(
                [
                    ActionObjectPermission(user_code.id, ActionPermission.READ, x)
                    for x in user_code.output_readers
                ]
            )

        code_link = LinkedObject.from_obj(user_code, server_uid=context.server.id)

        # Requests made on low side are synced, and have their status computed instead of set manually.
        if user_code.is_l0_deployment:
            status_change = SyncedUserCodeStatusChange(
                value=UserCodeStatus.APPROVED,
                linked_obj=user_code.status_link,
                linked_user_code=code_link,
            )
        else:
            status_change = UserCodeStatusChange(
                value=UserCodeStatus.APPROVED,
                linked_obj=user_code.status_link,
                linked_user_code=code_link,
            )
        changes = [status_change]

        request = SubmitRequest(changes=changes)
        method = context.server.get_service_method(RequestService.submit)
        result = method(context=context, request=request, reason=reason)

        # The Request service already returns either a SyftSuccess or SyftError
        return result

    def _get_or_submit_user_code(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
    ) -> Result[UserCode, str]:
        """
        - If the code is a UserCode, check if it exists and return
        - If the code is a SubmitUserCode and the same code hash exists, return the existing code
        - If the code is a SubmitUserCode and the code hash does not exist, submit the code
        """
        if isinstance(code, UserCode):
            # Get existing UserCode
            user_code_or_err = self.stash.get_by_uid(context.credentials, code.id)
            if user_code_or_err.is_err():
                return user_code_or_err
            user_code = user_code_or_err.ok()
            if user_code is None:
                return Err("UserCode not found on this server.")
            return Ok(user_code)
        else:  # code: SubmitUserCode
            # Submit new UserCode, or get existing UserCode with the same code hash
            user_code_or_err = self._submit(context, code, exists_ok=True)  # type: ignore
            return user_code_or_err

    @service_method(
        path="code.request_code_execution",
        name="request_code_execution",
        roles=GUEST_ROLE_LEVEL,
    )
    def request_code_execution(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
        reason: str | None = "",
    ) -> Request | SyftError:
        """Request Code execution on user code"""

        user_code_or_err = self._get_or_submit_user_code(context, code)
        if user_code_or_err.is_err():
            return SyftError(message=user_code_or_err.err())
        user_code = user_code_or_err.ok()

        result = self._request_code_execution(
            context,
            user_code,
            reason,
        )
        return result

    @service_method(path="code.get_all", name="get_all", roles=GUEST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[UserCode] | SyftError:
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="code.get_by_id", name="get_by_id", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> UserCode | SyftError:
        """Get a User Code Item"""
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            user_code = result.ok()
            if (
                user_code
                and user_code.input_policy_state
                and context.server is not None
            ):
                # TODO replace with LinkedObject Context
                user_code.server_uid = context.server.id
            return user_code
        return SyftError(message=result.err())

    @service_method(
        path="code.get_all_for_user",
        name="get_all_for_user",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_for_user(
        self, context: AuthedServiceContext
    ) -> SyftSuccess | SyftError:
        """Get All User Code Items for User's VerifyKey"""
        # TODO: replace with incoming user context and key
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    def update_code_state(
        self, context: AuthedServiceContext, code_item: UserCode
    ) -> SyftSuccess | SyftError:
        context = context.as_root_context()
        result = self.stash.update(context.credentials, code_item)
        if result.is_ok():
            return SyftSuccess(message="Code State Updated")
        return SyftError(message="Unable to Update Code State")

    def load_user_code(self, context: AuthedServiceContext) -> None:
        result = self.stash.get_all(credentials=context.credentials)
        if result.is_ok():
            user_code_items = result.ok()
            # Filter out UserCode items that are not updated to the latest versio
            user_code_items = [x for x in user_code_items if isinstance(x, UserCode)]
            load_approved_policy_code(user_code_items=user_code_items, context=context)

    def is_execution_allowed(
        self,
        code: UserCode,
        context: AuthedServiceContext,
        output_policy: OutputPolicy | None,
    ) -> bool | SyftSuccess | SyftError | SyftNotReady:
        status = code.get_status(context)
        if not status.approved:
            return status.get_status_message()
        # Check if the user has permission to execute the code.
        elif not (has_code_permission := self.has_code_permission(code, context)):
            return has_code_permission
        elif not code.is_output_policy_approved(context):
            return SyftError(message=f"Output policy not approved. {code}")

        policy_is_valid = output_policy is not None and output_policy._is_valid(context)
        if not policy_is_valid:
            return policy_is_valid
        else:
            return True

    def is_execution_on_owned_args_allowed(
        self, context: AuthedServiceContext
    ) -> bool | SyftError:
        if context.role == ServiceRole.ADMIN:
            return True

        user_service = context.server.get_service("userservice")
        current_user = user_service.get_current_user(context=context)
        return current_user.mock_execution_permission

    def keep_owned_kwargs(
        self, kwargs: dict[str, Any], context: AuthedServiceContext
    ) -> dict[str, Any] | SyftError:
        """Return only the kwargs that are owned by the user"""

        action_service = context.server.get_service("actionservice")

        mock_kwargs = {}
        for k, v in kwargs.items():
            if isinstance(v, UID):
                # Jobs have UID kwargs instead of ActionObject
                v = action_service.get(context, uid=v)
                if v.is_ok():
                    v = v.ok()
            if (
                isinstance(v, ActionObject)
                and v.syft_client_verify_key == context.credentials
            ):
                mock_kwargs[k] = v
        return mock_kwargs

    def is_execution_on_owned_args(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        passed_kwargs: dict[str, Any],
    ) -> bool:
        # Check if all kwargs are owned by the user
        all_kwargs_are_owned = len(
            self.keep_owned_kwargs(passed_kwargs, context)
        ) == len(passed_kwargs)
        if not all_kwargs_are_owned:
            return False

        # Check if the kwargs match the code signature
        code = self.stash.get_by_uid(context.credentials, user_code_id)
        if code.is_err():
            return False
        code = code.ok()

        # Skip the datasite and context kwargs, they are passed by the backend
        code_kwargs = set(code.signature.parameters.keys()) - {"datasite", "context"}

        passed_kwarg_keys = set(passed_kwargs.keys())
        return passed_kwarg_keys == code_kwargs

    @service_method(path="code.call", name="call", roles=GUEST_ROLE_LEVEL)
    def call(
        self, context: AuthedServiceContext, uid: UID, **kwargs: Any
    ) -> CachedSyftObject | ActionObject | SyftSuccess | SyftError:
        """Call a User Code Function"""
        kwargs.pop("result_id", None)
        result = self._call(context, uid, **kwargs)
        if result.is_err():
            return SyftError(message=result.err())
        else:
            return result.ok()

    def valid_worker_pool_for_context(
        self, context: AuthedServiceContext, user_code: UserCode
    ) -> bool:
        """This is a temporary fix that is needed until every function is always just ran as job"""
        # relative
        from ...server.server import get_default_worker_pool_name

        has_custom_worker_pool = (
            user_code.worker_pool_name is not None
        ) and user_code.worker_pool_name != get_default_worker_pool_name()
        if has_custom_worker_pool and context.is_blocking_api_call:
            return False
        else:
            return True

    def _call(
        self,
        context: AuthedServiceContext,
        uid: UID,
        result_id: UID | None = None,
        **kwargs: Any,
    ) -> Result[ActionObject, Err]:
        """Call a User Code Function"""
        try:
            code_result = self.stash.get_by_uid(context.credentials, uid=uid)
            if code_result.is_err():
                return code_result
            code: UserCode = code_result.ok()

            # Set Permissions
            if self.is_execution_on_owned_args(context, uid, kwargs):
                if self.is_execution_on_owned_args_allowed(context):
                    # handles the case: if we have 1 or more owned args and execution permission
                    # handles the case: if we have 0 owned args and execution permission
                    context.has_execute_permissions = True
                elif len(kwargs) == 0:
                    # handles the case: if we have 0 owned args and execution permission
                    pass
                else:
                    return Err(
                        "You do not have the permissions for mock execution, please contact the admin"
                    )
            override_execution_permission = (
                context.has_execute_permissions or context.role == ServiceRole.ADMIN
            )

            # Override permissions bypasses the cache, since we do not check in/out policies
            skip_fill_cache = override_execution_permission
            # We do not read from output policy cache if there are mock arguments
            skip_read_cache = len(self.keep_owned_kwargs(kwargs, context)) > 0

            # Extract ids from kwargs
            kwarg2id = map_kwargs_to_id(kwargs)

            input_policy = code.get_input_policy(context)
            # relative

            # Check output policy
            if not override_execution_permission:
                output_history = code.get_output_history(context=context)
                if isinstance(output_history, SyftError):
                    return Err(output_history.message)
                output_policy = code.get_output_policy(context)

                can_execute = output_policy and self.is_execution_allowed(
                    code=code,
                    context=context,
                    output_policy=output_policy,
                )
                if not can_execute:
                    # We check output policy only in l2 deployment.
                    # code is from low side (L0 setup)
                    status = code.get_status(context)
                    if not status.approved:
                        # return Err(
                        #     "Execution denied: Your code is waiting for approval"
                        # )
                        return Err(status.get_status_message().message)
                    is_valid = (
                        output_policy._is_valid(context) if output_policy else False
                    )
                    if not is_valid or code.is_l0_deployment:
                        if len(output_history) > 0 and not skip_read_cache:
                            last_executed_output = output_history[-1]
                            # Check if the inputs of the last executed output match
                            # against the current input
                            if (
                                input_policy is not None
                                and not last_executed_output.check_input_ids(
                                    kwargs=kwarg2id
                                )
                            ):
                                inp_policy_validation = input_policy._is_valid(
                                    context,
                                    usr_input_kwargs=kwarg2id,
                                    code_item_id=code.id,
                                )
                                if inp_policy_validation.is_err():
                                    return inp_policy_validation

                            result: Result[ActionObject, str] = resolve_outputs(
                                context=context,
                                output_ids=last_executed_output.output_ids,
                            )
                            if result.is_err():
                                return result

                            res = delist_if_single(result.ok())
                            output_policy_message = ""
                            if code.is_l2_deployment:
                                # Skip output policy warning in L0 setup;
                                # admin overrides policy checks.
                                output_policy_message = is_valid.message
                            return Ok(
                                CachedSyftObject(
                                    result=res,
                                    error_msg=output_policy_message,
                                )
                            )
                        else:
                            return cast(Err, is_valid.to_result())
                    return can_execute.to_result()  # type: ignore

            # Execute the code item
            if not self.valid_worker_pool_for_context(context, code):
                return Err(
                    value="You tried to run a syft function attached to a worker pool in blocking mode,"
                    "which is currently not supported. Run your function with `blocking=False` to run"
                    " as a job on your worker pool"
                )
            action_service: ActionService = context.server.get_service("actionservice")  # type: ignore
            result_action_object: Result[ActionObject | TwinObject, str] = (
                action_service._user_code_execute(
                    context, code, kwarg2id, result_id=result_id
                )
            )
            if result_action_object.is_err():
                return result_action_object
            else:
                result_action_object = result_action_object.ok()

            output_result = action_service.set_result_to_store(
                result_action_object, context, code.get_output_policy(context)
            )

            if output_result.is_err():
                return output_result
            result = output_result.ok()

            # Apply Output Policy to the results and update the OutputPolicyState

            # this currently only works for nested syft_functions
            # and admins executing on high side (TODO, decide if we want to increment counter)
            # always store_execution_output on l0 setup
            is_l0_request = context.role == ServiceRole.ADMIN and code.is_l0_deployment
            if not skip_fill_cache and output_policy is not None or is_l0_request:
                res = code.store_execution_output(
                    context=context,
                    outputs=result,
                    job_id=context.job_id,
                    input_ids=kwarg2id,
                )
                if isinstance(res, SyftError):
                    return Err(res.message)

            # output_policy.update_policy(context, result)
            # code.output_policy = output_policy
            # res = self.update_code_state(context, code)
            # print(res)

            has_result_read_permission = action_service.has_read_permission(
                context, result.id
            )

            if isinstance(result, TwinObject):
                if has_result_read_permission:
                    return Ok(result.private)
                else:
                    return Ok(result.mock)
            elif result.is_mock:
                return Ok(result)
            elif result.syft_action_data_type is Err:
                # result contains the error but the request was handled correctly
                return Ok(result)
            elif has_result_read_permission:
                return Ok(result)
            else:
                return Ok(result.as_empty())
        except Exception as e:
            # stdlib
            import traceback

            return Err(value=f"Failed to run. {e}, {traceback.format_exc()}")

    def has_code_permission(
        self, code_item: UserCode, context: AuthedServiceContext
    ) -> SyftSuccess | SyftError:
        if not (
            context.credentials == context.server.verify_key
            or context.credentials == code_item.user_verify_key
        ):
            return SyftError(
                message=f"Code Execution Permission: {context.credentials} denied"
            )
        return SyftSuccess(message="you have permission")

    @service_method(
        path="code.store_execution_output",
        name="store_execution_output",
        roles=GUEST_ROLE_LEVEL,
    )
    def store_execution_output(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        outputs: Any,
        input_ids: dict[str, UID] | None = None,
        job_id: UID | None = None,
    ) -> ExecutionOutput | SyftError:
        code_result = self.stash.get_by_uid(context.credentials, user_code_id)
        if code_result.is_err():
            return SyftError(message=code_result.err())

        is_admin = context.role == ServiceRole.ADMIN
        code: UserCode = code_result.ok()
        if not code.get_status(context).approved and not is_admin:
            return SyftError(message="This UserCode is not approved")

        res = code.store_execution_output(
            context=context,
            outputs=outputs,
            job_id=job_id,
            input_ids=input_ids,
        )
        return res


def resolve_outputs(
    context: AuthedServiceContext,
    output_ids: list[UID],
) -> Result[list[ActionObject], str]:
    # relative
    from ...service.action.action_object import TwinMode

    if isinstance(output_ids, list):
        if len(output_ids) == 0:
            return None
        outputs = []
        for output_id in output_ids:
            if context.server is not None:
                action_service = context.server.get_service("actionservice")
                result = action_service.get(
                    context, uid=output_id, twin_mode=TwinMode.PRIVATE
                )
                if result.is_err():
                    return result
                outputs.append(result.ok())
        return Ok(outputs)
    else:
        raise NotImplementedError


T = TypeVar("T")


def delist_if_single(result: list[T]) -> T | list[T]:
    if len(result) == 1:
        return result[0]
    return result


def map_kwargs_to_id(kwargs: dict[str, Any]) -> dict[str, Any]:
    # relative
    from ...types.twin_object import TwinObject
    from ..action.action_object import ActionObject
    from ..dataset.dataset import Asset

    filtered_kwargs = {}
    for k, v in kwargs.items():
        value = v
        if isinstance(v, ActionObject):
            value = v.id
        if isinstance(v, TwinObject):
            value = v.id
        if isinstance(v, Asset):
            value = v.action_id

        if not isinstance(value, UID):
            raise Exception(f"Input {k} must have a UID not {type(v)}")
        filtered_kwargs[k] = value
    return filtered_kwargs


TYPE_TO_SERVICE[UserCode] = UserCodeService
SERVICE_TO_TYPES[UserCodeService].update({UserCode})



# future
from __future__ import annotations

# stdlib
from collections.abc import Callable
from collections.abc import Iterable
from enum import Enum
import inspect
from io import BytesIO
import logging
from pathlib import Path
import sys
import threading
import time
import traceback
import types
from typing import Any
from typing import ClassVar
from typing import TYPE_CHECKING

# third party
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...client.api import APIRegistry
from ...client.api import SyftAPI
from ...client.api import SyftAPICall
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...serde.serialize import _serialize as serialize
from ...server.credentials import SyftVerifyKey
from ...service.blob_storage.util import can_upload_to_blob_storage
from ...service.response import SyftError
from ...service.response import SyftSuccess
from ...service.response import SyftWarning
from ...store.linked_obj import LinkedObject
from ...types.base import SyftBaseModel
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftBaseObject
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util.util import prompt_warning_message
from ..context import AuthedServiceContext
from ..response import SyftException
from ..service import from_api_or_context
from .action_data_empty import ActionDataEmpty
from .action_data_empty import ActionDataLink
from .action_data_empty import ObjectNotReady
from .action_permissions import ActionPermission
from .action_types import action_type_for_object
from .action_types import action_type_for_type
from .action_types import action_types

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # relative
    from ..sync.diff_state import AttrDiff

NoneType = type(None)


@serializable(canonical_name="TwinMode", version=1)
class TwinMode(Enum):
    NONE = 0
    PRIVATE = 1
    MOCK = 2


@serializable(canonical_name="ActionType", version=1)
class ActionType(Enum):
    GETATTRIBUTE = 1
    METHOD = 2
    SETATTRIBUTE = 4
    FUNCTION = 8
    CREATEOBJECT = 16
    SYFTFUNCTION = 32
    TWINAPI = 64


def repr_cls(c: Any) -> str:
    return f"{c.__module__}.{c.__name__}"


@serializable()
class Action(SyftObject):
    """Serializable Action object.

    Parameters:
        path: str
            The path of the Type of the remote object.
        op: str
            The method to be executed from the remote object.
        remote_self: Optional[LineageID]
            The extended UID of the SyftObject
        args: List[LineageID]
            `op` args
        kwargs: Dict[str, LineageID]
            `op` kwargs
        result_id: Optional[LineageID]
            Extended UID of the resulted SyftObject
    """

    __canonical_name__ = "Action"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__: ClassVar[list[str]] = []

    path: str | None = None
    op: str | None = None
    remote_self: LineageID | None = None
    args: list[LineageID]
    kwargs: dict[str, LineageID]
    result_id: LineageID = Field(default_factory=lambda: LineageID(UID()))
    action_type: ActionType | None = None
    create_object: SyftObject | None = None
    user_code_id: UID | None = None

    @field_validator("result_id", mode="before")
    @classmethod
    def make_result_id(cls, v: Any) -> LineageID:
        return v if isinstance(v, LineageID) else LineageID(v)

    @property
    def full_path(self) -> str:
        """Action path and operation"""
        return f"{self.path}.{self.op}"

    @property
    def job_display_name(self) -> str:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if self.user_code_id is not None and api is not None:
            user_code = api.services.code.get_by_id(self.user_code_id)
            return user_code.service_func_name
        else:
            return f"{self.path}.{self.op}"

    @property
    def syft_history_hash(self) -> int:
        """Create a unique hash for the operations applied on the object."""
        hashes = 0
        if self.remote_self:
            hashes += hash(self.remote_self.syft_history_hash)
        # 🔵 TODO: resolve this
        # if the object is ActionDataEmpty then the type might not be equal to the
        # real thing. This is the same issue with determining the result type from
        # a pointer operation in the past, so we should think about what we want here
        # hashes += hash(self.path)
        hashes += hash(self.op)
        for arg in self.args:
            hashes += hash(arg.syft_history_hash)
        for k, arg in self.kwargs.items():
            hashes += hash(k)
            hashes += hash(arg.syft_history_hash)
        return hashes

    @classmethod
    def syft_function_action_from_kwargs_and_id(
        cls, kwargs: dict[str, Any], user_code_id: UID
    ) -> Self:
        kwarg_ids = {}
        for k, v in kwargs.items():
            kwarg_ids[k] = LineageID(v)
        return cls(
            args=[],
            kwargs=kwarg_ids,
            result_id=LineageID(),
            action_type=ActionType.SYFTFUNCTION,
            user_code_id=user_code_id,
        )

    @classmethod
    def from_api_call(cls, api_call: SyftAPICall) -> Action:
        # relative
        from ..code.user_code_service import map_kwargs_to_id

        kwargs = api_call.kwargs
        kwargs.pop("communication_protocol", None)
        function_id = kwargs.pop("uid", None)
        kwargs = map_kwargs_to_id(kwargs)
        kwarg_ids = {}
        for k, v in kwargs.items():
            kwarg_ids[k] = LineageID(v)

        action = cls(
            args=[],
            kwargs=kwarg_ids,
            result_id=LineageID(),
            action_type=ActionType.SYFTFUNCTION,
            user_code_id=function_id,
        )
        return action

    @classmethod
    def from_api_endpoint_execution(cls: type[Self]) -> Action:
        return cls(
            args=[],
            kwargs={},
            result_id=LineageID(),
            action_type=ActionType.TWINAPI,
            user_code_id=None,
        )

    def __repr__(self) -> str:
        def repr_uid(_id: LineageID) -> str:
            return f"{str(_id)[:3]}..{str(_id)[-1]}"

        arg_repr = ", ".join([repr_uid(x) for x in self.args])
        kwargs_repr = ", ".join(
            [f"{key}={repr_uid(value)}" for key, value in self.kwargs.items()]
        )
        _coll_repr_ = (
            f"[{repr_uid(self.remote_self)}]" if self.remote_self is not None else ""
        )
        return (
            f"ActionObject {self.path}{_coll_repr_}.{self.op}({arg_repr},{kwargs_repr})"
        )


@serializable(canonical_name="ActionObjectPointer", version=1)
class ActionObjectPointer:
    pass


# Hooks
HOOK_ALWAYS = "ALWAYS"
HOOK_ON_POINTERS = "ON_POINTERS"

passthrough_attrs = [
    "__dict__",  # python
    "__class__",  # python
    "__repr_name__",  # python
    "__annotations__",  # python
    "_init_private_attributes",  # pydantic
    "__private_attributes__",  # pydantic
    "__config__",  # pydantic
    "__fields__",  # pydantic
    "__fields_set__",  # pydantic
    "__repr_str__",  # pydantic
    "__repr_args__",  # pydantic
    "__post_init__",  # syft
    "_get_api",  # syft
    "__validate_private_attrs__",  # syft
    "id",  # syft
    "created_date",  # syft
    "updated_date",  # syft
    "deleted_date",  # syft
    "to_mongo",  # syft 🟡 TODO 23: Add composeable / inheritable object passthrough attrs
    "__attr_searchable__",  # syft
    "__canonical_name__",  # syft
    "__version__",  # syft
    "__args__",  # pydantic
    "to_pointer",  # syft
    "to",  # syft
    "send",  # syft
    "_send",  # syft
    "_copy_and_set_values",  # pydantic
    "get_from",  # syft
    "get",  # syft
    "delete_data",  # syft
    "_save_to_blob_storage_",  # syft
    "syft_action_data",  # syft
    "syft_resolved",  # syft
    "syft_action_data_server_id",
    "server_uid",
    "migrate_to",  # syft
    "to_dict",  # syft
    "dict",  # syft
    "has_storage_permission",  # syft
    "_iter",  # pydantic
    "__exclude_fields__",  # pydantic
    "__include_fields__",  # pydantic
    "_calculate_keys",  # pydantic
    "_get_value",  # pydantic
    "__pydantic_validator__",  # pydantic
    "__class_vars__",  # pydantic
    "__private_attributes__",  # pydantic
    "__signature__",  # pydantic
    "__pydantic_complete__",  # pydantic
    "__pydantic_core_schema__",  # pydantic
    "__pydantic_custom_init__",  # pydantic
    "__pydantic_decorators__",  # pydantic
    "__pydantic_generic_metadata__",  # pydantic
    "__pydantic_parent_namespace__",  # pydantic
    "__pydantic_post_init__",  # pydantic
    "__pydantic_root_model__",  # pydantic
    "__pydantic_serializer__",  # pydantic
    "__pydantic_validator__",  # pydantic
    "__pydantic_extra__",  # pydantic
    "__pydantic_fields_set__",  # pydantic
    "__pydantic_private__",  # pydantic
    "model_config",  # pydantic
    "model_computed_fields",  # pydantic
    "model_extra",  # pydantic
    "model_fields",  # pydantic
    "model_fields_set",  # pydantic
    "model_construct",  # pydantic
    "model_copy",  # pydantic
    "model_dump",  # pydantic
    "model_dump_json",  # pydantic
    "model_json_schema",  # pydantic
    "model_parametrized_name",  # pydantic
    "model_post_init",  # pydantic
    "model_rebuild",  # pydantic
    "model_validate",  # pydantic
    "model_validate_json",  # pydantic
    "copy",  # pydantic
    "__sha256__",  # syft
    "__hash_exclude_attrs__",  # syft
    "__private_sync_attr_mocks__",  # syft
    "__exclude_sync_diff_attrs__",  # syft
    "__repr_attrs__",  # syft
    "get_sync_dependencies",
    "_data_repr",
    "syft_eq",  # syft
    "__table_coll_widths__",
    "_clear_cache",
    "_set_reprs",
]
dont_wrap_output_attrs = [
    "__repr__",
    "__str__",
    "__repr_attrs__",
    "_repr_html_",
    "_repr_markdown_",
    "_repr_latex_",
    "__array_struct__",
    "__array_prepare__",
    "__array_wrap__",
    "__bool__",
    "__len__",
    "syft_resolved",  # syft
    "server_uid",
    "syft_action_data_server_id",
    "__sha256__",
    "__hash_exclude_attrs__",
    "__exclude_sync_diff_attrs__",  # syft
    "__repr_attrs__",  # syft
    "get_sync_dependencies",  # syft
    "syft_eq",  # syft
    "__table_coll_widths__",
    "_clear_cache",
    "_set_reprs",
]
dont_make_side_effects = [
    "__repr_attrs__",
    "_repr_html_",
    "_repr_markdown_",
    "_repr_latex_",
    "__repr__",
    "__getitem__",
    "__setitem__",
    "__len__",
    "shape",
    "syft_resolved",  # syft
    "server_uid",
    "syft_action_data_server_id",
    "__sha256__",
    "__hash_exclude_attrs__",
    "__exclude_sync_diff_attrs__",  # syft
    "__repr_attrs__",
    "get_sync_dependencies",
    "syft_eq",  # syft
    "__table_coll_widths__",
    "_clear_cache",
    "_set_reprs",
]
action_data_empty_must_run = [
    "__repr__",
    "__str__",
]

methods_to_check_in_cache = [
    "_ipython_display_",
    "_repr_mimebundle_",
    "_repr_latex_",
    "_repr_javascript_",
    "_repr_html_",
    "_repr_jpeg_",
    "_repr_png_",
    "_repr_svg_",
    "_repr_pretty_",
    "_repr_pdf_",
    "_repr_json_",
    "_repr_markdown_",
]


class PreHookContext(SyftBaseObject):
    __canonical_name__ = "PreHookContext"
    __version__ = SYFT_OBJECT_VERSION_1

    """Hook context

    Parameters:
        obj: Any
            The ActionObject to use for the action
        op_name: str
            The method name to use for the action
        server_uid: Optional[UID]
            Optional Syft server UID
        result_id: Optional[Union[UID, LineageID]]
            Optional result Syft UID
        action: Optional[Action]
            The action generated by the current hook
    """

    obj: Any = None
    op_name: str
    server_uid: UID | None = None
    result_id: UID | LineageID | None = None
    result_twin_type: TwinMode | None = None
    action: Action | None = None
    action_type: ActionType | None = None


def make_action_side_effect(
    context: PreHookContext, *args: Any, **kwargs: Any
) -> Result[Ok[tuple[PreHookContext, tuple[Any, ...], dict[str, Any]]], Err[str]]:
    """Create a new action from context_op_name, and add it to the PreHookContext

    Parameters:
        context: PreHookContext
            PreHookContext object
        *args:
            Operation *args
        **kwargs
            Operation *kwargs
    Returns:
        - Ok[[Tuple[PreHookContext, Tuple[Any, ...], Dict[str, Any]]] on success
        - Err[str] on failure
    """
    try:
        action = context.obj.syft_make_action_with_self(
            op=context.op_name,
            args=args,
            kwargs=kwargs,
            action_type=context.action_type,
        )
        context.action = action
    except Exception as e:
        msg = "make_action_side_effect failed"
        logger.error(msg, exc_info=e)
        return Err(f"{msg} with {traceback.format_exc()}")

    return Ok((context, args, kwargs))


class TraceResultRegistry:
    __result_registry__: dict[int, TraceResult] = {}

    @classmethod
    def set_trace_result_for_current_thread(
        cls,
        client: SyftClient,
    ) -> None:
        cls.__result_registry__[threading.get_ident()] = TraceResult(
            client=client, is_tracing=True
        )

    @classmethod
    def get_trace_result_for_thread(cls) -> TraceResult | None:
        return cls.__result_registry__.get(threading.get_ident(), None)

    @classmethod
    def reset_result_for_thread(cls) -> None:
        if threading.get_ident() in cls.__result_registry__:
            del cls.__result_registry__[threading.get_ident()]

    @classmethod
    def current_thread_is_tracing(cls) -> bool:
        trace_result = cls.get_trace_result_for_thread()
        if trace_result is None:
            return False
        else:
            return trace_result.is_tracing


class TraceResult(SyftBaseModel):
    result: list = []
    client: SyftClient
    is_tracing: bool = False


def trace_action_side_effect(
    context: PreHookContext, *args: Any, **kwargs: Any
) -> Result[Ok[tuple[PreHookContext, tuple[Any, ...], dict[str, Any]]], Err[str]]:
    action = context.action
    if action is not None and TraceResultRegistry.current_thread_is_tracing():
        trace_result = TraceResultRegistry.get_trace_result_for_thread()
        trace_result.result += [action]  # type: ignore
    return Ok((context, args, kwargs))


def convert_to_pointers(
    api: SyftAPI,
    server_uid: UID | None = None,
    args: list | None = None,
    kwargs: dict | None = None,
) -> tuple[list, dict]:
    # relative
    from ..dataset.dataset import Asset

    def process_arg(arg: ActionObject | Asset | UID | Any) -> Any:
        if (
            not isinstance(arg, ActionObject | Asset | UID)
            and api.signing_key is not None  # type: ignore[unreachable]
        ):
            arg = ActionObject.from_obj(  # type: ignore[unreachable]
                syft_action_data=arg,
                syft_client_verify_key=api.signing_key.verify_key,
                syft_server_location=api.server_uid,
            )
            arg.syft_server_uid = server_uid
            r = arg._save_to_blob_storage()
            if isinstance(r, SyftError):
                print(r.message)
            if isinstance(r, SyftWarning):
                logger.debug(r.message)
            arg = api.services.action.set(arg)
        return arg

    arg_list = [process_arg(arg) for arg in args] if args else []
    kwarg_dict = {k: process_arg(v) for k, v in kwargs.items()} if kwargs else {}

    return arg_list, kwarg_dict


def send_action_side_effect(
    context: PreHookContext, *args: Any, **kwargs: Any
) -> Result[Ok[tuple[PreHookContext, tuple[Any, ...], dict[str, Any]]], Err[str]]:
    """Create a new action from the context.op_name, and execute it on the remote server."""
    try:
        if context.action is None:
            result = make_action_side_effect(context, *args, **kwargs)
            if result.is_err():
                raise RuntimeError(result.err())

            context, _, _ = result.ok()

        action_result = context.obj.syft_execute_action(context.action, sync=True)

        if not isinstance(action_result, ActionObject):
            raise RuntimeError(f"Got back unexpected response : {action_result}")
        else:
            context.server_uid = action_result.syft_server_uid
            context.result_id = action_result.id
            context.result_twin_type = action_result.syft_twin_type
    except Exception as e:
        return Err(
            f"send_action_side_effect failed with {e}\n {traceback.format_exc()}"
        )
    return Ok((context, args, kwargs))


def propagate_server_uid(
    context: PreHookContext, op: str, result: Any
) -> Result[Ok[Any], Err[str]]:
    """Patch the result to include the syft_server_uid

    Parameters:
        context: PreHookContext
            PreHookContext object
        op: str
            Which operation was executed
        result: Any
            The result to patch
    Returns:
        - Ok[[result] on success
        - Err[str] on failure
    """
    if context.op_name in dont_make_side_effects or not hasattr(
        context.obj, "syft_server_uid"
    ):
        return Ok(result)

    try:
        syft_server_uid = getattr(context.obj, "syft_server_uid", None)
        if syft_server_uid is None:
            raise RuntimeError(
                "Can't proagate server_uid because parent doesnt have one"
            )

        if op not in context.obj._syft_dont_wrap_attrs():
            if hasattr(result, "syft_server_uid"):
                result.syft_server_uid = syft_server_uid
        else:
            raise RuntimeError("dont propogate server_uid because output isnt wrapped")
    except Exception:
        return Err(f"propagate_server_uid failed with {traceback.format_exc()}")

    return Ok(result)


def debox_args_and_kwargs(args: Any, kwargs: Any) -> tuple[Any, Any]:
    filtered_args = []
    filtered_kwargs = {}
    for a in args:
        value = a
        if hasattr(value, "syft_action_data"):
            value = value.syft_action_data
        filtered_args.append(value)

    for k, a in kwargs.items():
        value = a
        if hasattr(value, "syft_action_data"):
            value = value.syft_action_data
        filtered_kwargs[k] = a

    return tuple(filtered_args), filtered_kwargs


BASE_PASSTHROUGH_ATTRS: list[str] = [
    "is_mock",
    "is_real",
    "is_twin",
    "is_pointer",
    "request",
    "__repr__",
    "_repr_markdown_",
    "syft_twin_type",
    "_repr_debug_",
    "as_empty",
    "get",
    "is_link",
    "wait",
    "_save_to_blob_storage",
    "_save_to_blob_storage_",
    "syft_action_data",
    "__check_action_data",
    "as_empty_data",
    "_set_obj_location_",
    "syft_action_data_cache",
    "reload_cache",
    "syft_resolved",
    "refresh_object",
    "syft_action_data_server_id",
    "server_uid",
    "__sha256__",
    "__hash_exclude_attrs__",
    "__hash__",
    "create_shareable_sync_copy",
    "_has_private_sync_attrs",
    "__exclude_sync_diff_attrs__",
    "__repr_attrs__",
    "get_sync_dependencies",
    "_data_repr",
    "syft_eq",
    "__table_coll_widths__",
    "_clear_cache",
    "_set_reprs",
]


def truncate_str(string: str, length: int = 100) -> str:
    stringlen = len(string)
    if stringlen > length:
        n_hidden = stringlen - length
        string = f"{string[:length]}... ({n_hidden} characters hidden)"
    return string


@serializable(without=["syft_pre_hooks__", "syft_post_hooks__"])
class ActionObject(SyncableSyftObject):
    """Action object for remote execution."""

    __canonical_name__ = "ActionObject"
    __version__ = SYFT_OBJECT_VERSION_1
    __private_sync_attr_mocks__: ClassVar[dict[str, Any]] = {
        "syft_action_data_cache": None,
        "syft_blob_storage_entry_id": None,
    }

    __attr_searchable__: list[str] = []  # type: ignore[misc]
    syft_action_data_cache: Any | None = None
    syft_blob_storage_entry_id: UID | None = None
    syft_pointer_type: ClassVar[type[ActionObjectPointer]]

    # Help with calculating history hash for code verification
    syft_parent_hashes: int | list[int] | None = None
    syft_parent_op: str | None = None
    syft_parent_args: Any | None = None
    syft_parent_kwargs: Any | None = None
    syft_history_hash: int | None = None
    syft_internal_type: ClassVar[type[Any]]
    syft_server_uid: UID | None = None
    syft_pre_hooks__: dict[str, list] = {}
    syft_post_hooks__: dict[str, list] = {}
    syft_twin_type: TwinMode = TwinMode.NONE
    syft_passthrough_attrs: list[str] = BASE_PASSTHROUGH_ATTRS
    syft_action_data_type: type | None = None
    syft_action_data_repr_: str | None = None
    syft_action_data_str_: str | None = None
    syft_has_bool_attr: bool | None = None
    syft_resolve_data: bool | None = None
    syft_created_at: DateTime | None = None
    syft_resolved: bool = True
    syft_action_data_server_id: UID | None = None
    syft_action_saved_to_blob_store: bool = True
    # syft_dont_wrap_attrs = ["shape"]

    def syft_get_diffs(self, ext_obj: Any) -> list[AttrDiff]:
        # relative
        from ...service.sync.diff_state import AttrDiff

        diff_attrs = []

        # Sanity check
        if ext_obj.id != self.id:
            raise Exception("Not the same id for low side and high side requests")

        low_data = ext_obj.syft_action_data
        high_data = self.syft_action_data

        try:
            cmp = low_data != high_data
            if isinstance(cmp, Iterable):
                cmp = all(cmp)
        except Exception:
            cmp = False

        if cmp:
            diff_attr = AttrDiff(
                attr_name="syft_action_data", low_attr=low_data, high_attr=high_data
            )
            diff_attrs.append(diff_attr)
        return diff_attrs

    def _set_obj_location_(self, server_uid: UID, credentials: SyftVerifyKey) -> None:
        self.syft_server_location = server_uid
        self.syft_client_verify_key = credentials
        if self.syft_action_data_server_id is None:
            self.syft_action_data_server_id = server_uid

    @property
    def syft_action_data(self) -> Any:
        if self.syft_blob_storage_entry_id and self.syft_created_at:
            res = self.reload_cache()
            if isinstance(res, SyftError):
                print(res)
        return self.syft_action_data_cache

    def reload_cache(self) -> SyftError | None:
        # If ActionDataEmpty then try to fetch it from store.
        if isinstance(self.syft_action_data_cache, ActionDataEmpty):
            blob_storage_read_method = from_api_or_context(
                func_or_path="blob_storage.read",
                syft_server_location=self.syft_server_location,
                syft_client_verify_key=self.syft_client_verify_key,
            )

            if blob_storage_read_method is not None:
                blob_retrieval_object = blob_storage_read_method(
                    uid=self.syft_blob_storage_entry_id
                )
                if isinstance(blob_retrieval_object, SyftError):
                    logger.error(
                        f"Could not fetch actionobject data: {blob_retrieval_object}"
                    )
                    return blob_retrieval_object
                # relative
                from ...store.blob_storage import BlobRetrieval

                if isinstance(blob_retrieval_object, SyftError):
                    return blob_retrieval_object
                elif isinstance(blob_retrieval_object, BlobRetrieval):
                    # TODO: This change is temporary to for gateway to be compatible with the new blob storage
                    self.syft_action_data_cache = blob_retrieval_object.read()
                    self.syft_action_data_type = type(self.syft_action_data)
                    return None
                else:
                    # In the case of gateway, we directly receive the actual object
                    # TODO: The ideal solution would be to stream the object from the datasite through the gateway
                    # Currently , we are just passing the object as it is, which would be fixed later.
                    self.syft_action_data_cache = blob_retrieval_object
                    self.syft_action_data_type = type(self.syft_action_data)
                    return None
            else:
                return SyftError(
                    message="Could not reload cache, could not get read method"
                )

        return None

    def _save_to_blob_storage_(self, data: Any) -> SyftError | SyftWarning | None:
        # relative
        from ...types.blob_storage import BlobFile
        from ...types.blob_storage import CreateBlobStorageEntry

        if not isinstance(data, ActionDataEmpty):
            if isinstance(data, BlobFile):
                if not data.uploaded:
                    api = APIRegistry.api_for(
                        self.syft_server_location, self.syft_client_verify_key
                    )
                    data._upload_to_blobstorage_from_api(api)
            else:
                get_metadata = from_api_or_context(
                    func_or_path="metadata.get_metadata",
                    syft_server_location=self.syft_server_location,
                    syft_client_verify_key=self.syft_client_verify_key,
                )
                if get_metadata is not None and not can_upload_to_blob_storage(
                    data, get_metadata()
                ):
                    self.syft_action_saved_to_blob_store = False
                    return SyftWarning(
                        message=f"The action object {self.id} was not saved to "
                        f"the blob store but to memory cache since it is small."
                    )
                serialized = serialize(data, to_bytes=True)
                size = sys.getsizeof(serialized)
                storage_entry = CreateBlobStorageEntry.from_obj(data, file_size=size)

                if not TraceResultRegistry.current_thread_is_tracing():
                    self.syft_action_data_cache = self.as_empty_data()
                if self.syft_blob_storage_entry_id is not None:
                    # TODO: check if it already exists
                    storage_entry.id = self.syft_blob_storage_entry_id
                allocate_method = from_api_or_context(
                    func_or_path="blob_storage.allocate",
                    syft_server_location=self.syft_server_location,
                    syft_client_verify_key=self.syft_client_verify_key,
                )
                if allocate_method is not None:
                    blob_deposit_object = allocate_method(storage_entry)
                    if isinstance(blob_deposit_object, SyftError):
                        return blob_deposit_object

                    result = blob_deposit_object.write(BytesIO(serialized))
                    if isinstance(result, SyftError):
                        return result

                    self.syft_blob_storage_entry_id = (
                        blob_deposit_object.blob_storage_entry_id
                    )
                else:
                    logger.warn("cannot save to blob storage. allocate_method=None")

            self.syft_action_data_type = type(data)
            self._set_reprs(data)
            self.syft_has_bool_attr = hasattr(data, "__bool__")
        else:
            logger.debug(
                "skipping writing action object to store, passed data was empty."
            )

        self.syft_action_data_cache = data

        return None

    def _save_to_blob_storage(
        self, allow_empty: bool = False
    ) -> SyftError | SyftSuccess | SyftWarning:
        data = self.syft_action_data
        if isinstance(data, SyftError):
            return data

        if isinstance(data, ActionDataEmpty):
            return SyftError(
                message=f"cannot store empty object {self.id} to the blob storage"
            )

        try:
            result = self._save_to_blob_storage_(data)
            if isinstance(result, SyftError | SyftWarning):
                return result
            if not TraceResultRegistry.current_thread_is_tracing():
                self._clear_cache()
            return SyftSuccess(
                message=f"Saved action object {self.id} to the blob store"
            )
        except Exception as e:
            raise e

    def _clear_cache(self) -> None:
        self.syft_action_data_cache = self.as_empty_data()

    def _set_reprs(self, data: any) -> None:
        if inspect.isclass(data):
            self.syft_action_data_repr_ = truncate_str(repr_cls(data))
        else:
            self.syft_action_data_repr_ = truncate_str(
                data._repr_markdown_()
                if hasattr(data, "_repr_markdown_")
                else data.__repr__()
            )
        self.syft_action_data_str_ = truncate_str(str(data))

    @property
    def is_pointer(self) -> bool:
        return self.syft_server_uid is not None

    @property
    def syft_lineage_id(self) -> LineageID:
        """Compute the LineageID of the ActionObject, using the `id` and the `syft_history_hash` members"""
        return LineageID(self.id, self.syft_history_hash)

    model_config = ConfigDict(validate_assignment=True)

    @model_validator(mode="before")
    @classmethod
    def __check_action_data(cls, values: dict) -> dict:
        v = values.get("syft_action_data_cache")
        if values.get("syft_action_data_type", None) is None:
            values["syft_action_data_type"] = type(v)
        if not isinstance(v, ActionDataEmpty):
            if inspect.isclass(v):
                values["syft_action_data_repr_"] = truncate_str(repr_cls(v))
            else:
                values["syft_action_data_repr_"] = truncate_str(
                    v._repr_markdown_()
                    if v is not None and hasattr(v, "_repr_markdown_")
                    else v.__repr__()
                )
            values["syft_action_data_str_"] = truncate_str(str(v))
            values["syft_has_bool_attr"] = hasattr(v, "__bool__")
        return values

    @property
    def is_mock(self) -> bool:
        return self.syft_twin_type == TwinMode.MOCK

    @property
    def is_real(self) -> bool:
        return self.syft_twin_type == TwinMode.PRIVATE

    @property
    def is_twin(self) -> bool:
        return self.syft_twin_type != TwinMode.NONE

    def syft_point_to(self, server_uid: UID) -> ActionObject:
        """Set the syft_server_uid, used in the post hooks"""
        self.syft_server_uid = server_uid
        return self

    def syft_get_property(self, obj: Any, method: str) -> Any:
        klass_method = getattr(type(obj), method, None)
        if klass_method is None:
            raise Exception(f"{type(obj)} has no {method} attribute")
        return klass_method.__get__(obj)

    def syft_is_property(self, obj: Any, method: str) -> bool:
        klass_method = getattr(type(obj), method, None)
        return isinstance(klass_method, property) or inspect.isdatadescriptor(
            klass_method
        )

    def syft_eq(self, ext_obj: Self | None) -> bool:
        if ext_obj is None:
            return False
        return self.id.id == ext_obj.id.id

    def syft_execute_action(
        self, action: Action, sync: bool = True
    ) -> ActionObjectPointer:
        """Execute a remote action

        Parameters:
            action: Action
                Which action to execute
            sync: bool
                Run sync/async

        Returns:
            ActionObjectPointer
        """
        if self.syft_server_uid is None:
            raise SyftException("Pointers can't execute without a server_uid.")

        # relative
        from ...client.api import APIRegistry
        from ...client.api import SyftAPICall

        api = APIRegistry.api_for(
            server_uid=self.syft_server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(f"api is None. You must login to {self.syft_server_uid}")
        kwargs = {"action": action}
        api_call = SyftAPICall(
            server_uid=self.syft_server_uid,
            path="action.execute",
            args=[],
            kwargs=kwargs,
        )
        return api.make_call(api_call)

    def request(self, client: SyftClient) -> Any | SyftError:
        # relative
        from ..request.request import ActionStoreChange
        from ..request.request import SubmitRequest

        action_object_link = LinkedObject.from_obj(
            self, server_uid=self.syft_server_uid
        )
        permission_change = ActionStoreChange(
            linked_obj=action_object_link, apply_permission_type=ActionPermission.READ
        )
        if client.credentials is None:
            return SyftError(f"{client} has no signing key")
        submit_request = SubmitRequest(
            changes=[permission_change],
            requesting_user_verify_key=client.credentials.verify_key,
        )
        return client.api.services.request.submit(submit_request)

    def _syft_try_to_save_to_store(self, obj: SyftObject) -> None:
        if self.syft_server_uid is None or self.syft_client_verify_key is None:
            return
        elif obj.syft_server_uid is not None:
            return

        if obj.syft_blob_storage_entry_id is not None:
            return
        # TODO fix: the APIRegistry often gets the wrong client
        # if you have 2 clients in memory
        # therefore the following happens if you call a method
        # with a pointer to a twin (mock version)
        # 1) it gets the wrong credentials
        # 2) it uses the mock version to overwrite the real version
        # 3) it shouldnt send in the first place as it already exists

        # relative
        from ...client.api import APIRegistry

        if obj.syft_server_location is None:
            obj.syft_server_location = obj.syft_server_uid

        action = Action(
            path="",
            op="",
            remote_self=None,
            result_id=obj.id,
            args=[],
            kwargs={},
            action_type=ActionType.CREATEOBJECT,
            create_object=obj,
        )

        if TraceResultRegistry.current_thread_is_tracing():
            trace_result = TraceResultRegistry.get_trace_result_for_thread()
            trace_result.result += [action]  # type: ignore

        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            print(
                f"failed saving {obj} to blob storage, api is None. You must login to {self.syft_server_location}"
            )
            return
        else:
            obj._set_obj_location_(api.server_uid, api.signing_key.verify_key)  # type: ignore[union-attr]

        res = api.services.action.execute(action)
        if isinstance(res, SyftError):
            print(f"Failed to to store (arg) {obj} to store, {res}")

    def _syft_prepare_obj_uid(self, obj: Any) -> LineageID:
        # We got the UID
        if isinstance(obj, UID | LineageID):
            return LineageID(obj.id)

        # We got the ActionObjectPointer
        if isinstance(obj, ActionObjectPointer):
            return obj.syft_lineage_id

        # We got the ActionObject. We need to save it in the store.
        if isinstance(obj, ActionObject):
            self._syft_try_to_save_to_store(obj)
            return obj.syft_lineage_id

        # We got a raw object. We need to create the ActionObject from scratch and save it in the store.
        act_obj = ActionObject.from_obj(
            obj,
            syft_client_verify_key=self.syft_client_verify_key,
            syft_server_location=self.syft_server_location,
        )

        self._syft_try_to_save_to_store(act_obj)

        return act_obj.syft_lineage_id

    def syft_make_action(
        self,
        path: str,
        op: str,
        remote_self: UID | LineageID | None = None,
        args: (
            list[UID | LineageID | ActionObjectPointer | ActionObject | Any] | None
        ) = None,
        kwargs: (
            dict[str, UID | LineageID | ActionObjectPointer | ActionObject | Any] | None
        ) = None,
        action_type: ActionType | None = None,
    ) -> Action:
        """Generate new action from the information

        Parameters:
            path: str
                The path of the Type of the remote object.
            op: str
                The method to be executed from the remote object.
            remote_self: Optional[Union[UID, LineageID]]
                The extended UID of the SyftObject
            args: Optional[List[Union[UID, LineageID, ActionObjectPointer, ActionObject]]]
                `op` args
            kwargs: Optional[Dict[str, Union[UID, LineageID, ActionObjectPointer, ActionObject]]]
                `op` kwargs
        Returns:
            Action object

        Raises:
            ValueError: For invalid args or kwargs
            PydanticValidationError: For args and kwargs
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        arg_ids = [self._syft_prepare_obj_uid(obj) for obj in args]

        kwarg_ids = {k: self._syft_prepare_obj_uid(obj) for k, obj in kwargs.items()}

        action = Action(
            path=path,
            op=op,
            remote_self=LineageID(remote_self),
            args=arg_ids,
            kwargs=kwarg_ids,
            action_type=action_type,
        )
        return action

    def syft_make_action_with_self(
        self,
        op: str,
        args: list[UID | ActionObjectPointer] | None = None,
        kwargs: dict[str, UID | ActionObjectPointer] | None = None,
        action_type: ActionType | None = None,
    ) -> Action:
        """Generate new method action from the current object.

        Parameters:
            op: str
                The method to be executed from the remote object.
            args: List[LineageID]
                `op` args
            kwargs: Dict[str, LineageID]
                `op` kwargs
        Returns:
            Action object

        Raises:
            ValueError: For invalid args or kwargs
            PydanticValidationError: For args and kwargs
        """
        path = self.syft_get_path()
        return self.syft_make_action(
            path=path,
            op=op,
            remote_self=self.syft_lineage_id,
            args=args,
            kwargs=kwargs,
            action_type=action_type,
        )

    def get_sync_dependencies(
        self, context: AuthedServiceContext, **kwargs: dict
    ) -> list[UID]:  # type: ignore
        # relative
        from ..job.job_stash import Job

        job_service = context.server.get_service("jobservice")  # type: ignore
        job: Job | None | SyftError = job_service.get_by_result_id(context, self.id.id)  # type: ignore
        if isinstance(job, SyftError):
            return job
        elif job is not None:
            return [job.id]
        else:
            return []

    def syft_get_path(self) -> str:
        """Get the type path of the underlying object"""
        if (
            isinstance(self, AnyActionObject)
            and self.syft_internal_type
            and self.syft_action_data_type is not None
        ):
            # avoids AnyActionObject errors
            return f"{self.syft_action_data_type.__name__}"
        return f"{type(self).__name__}"

    def syft_remote_method(
        self,
        op: str,
    ) -> Callable:
        """Generate a Callable object for remote calls.

        Parameters:
            op: str
                he method to be executed from the remote object.

        Returns:
            A function
        """

        def wrapper(
            *args: list[UID | ActionObjectPointer] | None,
            **kwargs: dict[str, UID | ActionObjectPointer] | None,
        ) -> Action:
            return self.syft_make_action_with_self(op=op, args=args, kwargs=kwargs)

        return wrapper

    def send(self, client: SyftClient) -> Any:
        return self._send(
            server_uid=client.id,
            verify_key=client.verify_key,
            add_storage_permission=True,
        )

    def _send(
        self,
        server_uid: UID,
        verify_key: SyftVerifyKey,
        add_storage_permission: bool = True,
    ) -> Self | SyftError:
        self._set_obj_location_(server_uid, verify_key)

        blob_storage_res = self._save_to_blob_storage()
        if isinstance(blob_storage_res, SyftError):
            return blob_storage_res

        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        if isinstance(blob_storage_res, SyftWarning):
            logger.debug(blob_storage_res.message)
        res = api.services.action.set(
            self,
            add_storage_permission=add_storage_permission,
        )
        if isinstance(res, ActionObject):
            self.syft_created_at = res.syft_created_at
        return res

    def get_from(self, client: SyftClient) -> Any:
        """Get the object from a Syft Client"""
        res = client.api.services.action.get(self.id)
        if not isinstance(res, ActionObject):
            return SyftError(message=f"{res}")
        else:
            return res.syft_action_data

    def refresh_object(self, resolve_nested: bool = True) -> ActionObject | SyftError:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"api is None. You must login to {self.syft_server_location}"
            )

        res = api.services.action.get(self.id, resolve_nested=resolve_nested)
        return res

    def has_storage_permission(self) -> bool:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )

        if api is None:
            return False

        return api.services.action.has_storage_permission(self.id)

    def get(self, block: bool = False) -> Any:
        """Get the object from a Syft Client"""
        # relative

        if block:
            self.wait()

        res = self.refresh_object()
        if not isinstance(res, ActionObject):
            return SyftError(message=f"{res}")  # type: ignore
        elif issubclass(res.syft_action_data_type, Err):
            return SyftError(message=f"{res.syft_action_data.err()}")
        else:
            if not self.has_storage_permission():
                prompt_warning_message(
                    message="This is a placeholder object, the real data lives on a different server and is not synced."
                )
            nested_res = res.syft_action_data
            if isinstance(nested_res, ActionObject):
                nested_res.syft_server_location = res.syft_server_location
                nested_res.syft_client_verify_key = res.syft_client_verify_key
            return nested_res

    def as_empty(self) -> ActionObject:
        id = self.id

        if isinstance(id, LineageID):
            id = id.id

        res = ActionObject.empty(
            self.syft_internal_type,
            id,
            self.syft_lineage_id,
            self.syft_resolved,
            syft_blob_storage_entry_id=self.syft_blob_storage_entry_id,
        )
        if isinstance(self.id, LineageID):
            res.id = self.id

        return res

    def create_shareable_sync_copy(self, mock: bool) -> ActionObject:
        if mock:
            res = self.as_empty()
            for k, v in self.__private_sync_attr_mocks__.items():
                setattr(res, k, v)
            return res
        return self

    @staticmethod
    def from_path(
        path: str | Path,
        id: UID | None = None,
        syft_lineage_id: LineageID | None = None,
        syft_client_verify_key: SyftVerifyKey | None = None,
        syft_server_location: UID | None = None,
    ) -> ActionObject:
        """Create an Action Object from a file."""
        # relative
        from ...types.blob_storage import BlobFile

        if id is not None and syft_lineage_id is not None and id != syft_lineage_id.id:
            raise ValueError("UID and LineageID should match")

        _path = path if isinstance(path, Path) else Path(path)
        syft_action_data = BlobFile(path=_path, file_name=_path.name)

        action_type = action_type_for_object(syft_action_data)

        action_object = action_type(syft_action_data_cache=syft_action_data)

        if id is not None:
            action_object.id = id

        if syft_client_verify_key is not None:
            action_object.syft_client_verify_key = syft_client_verify_key

        if syft_server_location is not None:
            action_object.syft_server_location = syft_server_location

        if syft_lineage_id is not None:
            action_object.id = syft_lineage_id.id
            action_object.syft_history_hash = syft_lineage_id.syft_history_hash
        elif id is not None:
            action_object.syft_history_hash = hash(id)

        return action_object

    @staticmethod
    def from_obj(
        syft_action_data: Any,
        id: UID | None = None,
        syft_lineage_id: LineageID | None = None,
        syft_client_verify_key: SyftVerifyKey | None = None,
        syft_server_location: UID | None = None,
        syft_resolved: bool | None = True,
        data_server_id: UID | None = None,
        syft_blob_storage_entry_id: UID | None = None,
    ) -> ActionObject:
        """Create an ActionObject from an existing object.

        Parameters:
            syft_action_data: Any
                The object to be converted to a Syft ActionObject
            id: Optional[UID]
                Which ID to use for the ActionObject. Optional
            syft_lineage_id: Optional[LineageID]
                Which LineageID to use for the ActionObject. Optional
        """
        if id is not None and syft_lineage_id is not None and id != syft_lineage_id.id:
            raise ValueError("UID and LineageID should match")

        action_type = action_type_for_object(syft_action_data)
        action_object = action_type(syft_action_data_cache=syft_action_data)
        action_object.syft_blob_storage_entry_id = syft_blob_storage_entry_id
        action_object.syft_action_data_server_id = data_server_id
        action_object.syft_resolved = syft_resolved

        if id is not None:
            action_object.id = id

        if syft_client_verify_key is not None:
            action_object.syft_client_verify_key = syft_client_verify_key

        if syft_server_location is not None:
            action_object.syft_server_location = syft_server_location

        if syft_lineage_id is not None:
            action_object.id = syft_lineage_id.id
            action_object.syft_history_hash = syft_lineage_id.syft_history_hash
        elif id is not None:
            action_object.syft_history_hash = hash(id)

        return action_object

    @classmethod
    def add_trace_hook(cls) -> bool:
        return True
        # if trace_action_side_effect not in self.syft_pre_hooks__[HOOK_ALWAYS]:
        #     self.syft_pre_hooks__[HOOK_ALWAYS].append(trace_action_side_effect)

    @classmethod
    def remove_trace_hook(cls) -> bool:
        return True
        # self.syft_pre_hooks__[HOOK_ALWAYS].pop(trace_action_side_effct, None)

    def as_empty_data(self) -> ActionDataEmpty:
        return ActionDataEmpty(syft_internal_type=self.syft_internal_type)

    def wait(self, timeout: int | None = None) -> ActionObject | SyftError:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if isinstance(self.id, LineageID):
            obj_id = self.id.id
        else:
            obj_id = self.id

        counter = 0
        while api:
            obj_resolved: bool | str = api.services.action.is_resolved(obj_id)
            if isinstance(obj_resolved, str):
                return SyftError(message=obj_resolved)
            if obj_resolved:
                break
            if not obj_resolved:
                time.sleep(1)
                if timeout is not None:
                    counter += 1
                    if counter > timeout:
                        return SyftError(message="Reached Timeout!")

        return self

    @staticmethod
    def link(
        result_id: UID,
        pointer_id: UID | None = None,
    ) -> ActionObject:
        link = ActionDataLink(action_object_id=pointer_id)
        res = ActionObject.from_obj(
            id=result_id,
            syft_action_data=link,
        )
        return res

    @staticmethod
    def obj_not_ready(
        id: UID,
    ) -> ActionObject:
        inner_obj = ObjectNotReady(obj_id=id)

        res = ActionObject.from_obj(
            id=id,
            syft_action_data=inner_obj,
        )
        return res

    @classmethod
    def empty(
        # TODO: fix the mypy issue
        cls,
        syft_internal_type: type[Any] | None = None,
        id: UID | None = None,
        syft_lineage_id: LineageID | None = None,
        syft_resolved: bool | None = True,
        data_server_id: UID | None = None,
        syft_blob_storage_entry_id: UID | None = None,
    ) -> Self:
        """Create an ActionObject from a type, using a ActionDataEmpty object

        Parameters:
            syft_internal_type: Type
                The Type for which to create a ActionDataEmpty object
            id: Optional[UID]
                Which ID to use for the ActionObject. Optional
            syft_lineage_id: Optional[LineageID]
                Which LineageID to use for the ActionObject. Optional
        """

        syft_internal_type = (
            type(None) if syft_internal_type is None else syft_internal_type
        )
        empty = ActionDataEmpty(syft_internal_type=syft_internal_type)
        res = cls.from_obj(
            id=id,
            syft_lineage_id=syft_lineage_id,
            syft_action_data=empty,
            syft_resolved=syft_resolved,
            data_server_id=data_server_id,
            syft_blob_storage_entry_id=syft_blob_storage_entry_id,
        )
        res.__dict__["syft_internal_type"] = syft_internal_type
        return res

    def __post_init__(self) -> None:
        """Add pre/post hooks."""
        if HOOK_ALWAYS not in self.syft_pre_hooks__:
            self.syft_pre_hooks__[HOOK_ALWAYS] = []

        if HOOK_ON_POINTERS not in self.syft_pre_hooks__:
            self.syft_pre_hooks__[HOOK_ON_POINTERS] = []

        if HOOK_ALWAYS not in self.syft_post_hooks__:
            self.syft_post_hooks__[HOOK_ALWAYS] = []

        if HOOK_ON_POINTERS not in self.syft_post_hooks__:
            self.syft_post_hooks__[HOOK_ON_POINTERS] = []

        api = APIRegistry.api_for(
            self.syft_server_location, self.syft_client_verify_key
        )
        eager_execution_enabled = (
            api is not None
            and api.metadata is not None
            and api.metadata.eager_execution_enabled
        )

        self._syft_add_pre_hooks__(eager_execution_enabled)
        self._syft_add_post_hooks__(eager_execution_enabled)

        if isinstance(self.syft_action_data_type, ActionObject):
            raise Exception("Nested ActionObjects", self.syft_action_data_repr_)

        self.syft_history_hash = hash(self.id)

    def _syft_add_pre_hooks__(self, eager_execution: bool) -> None:
        """
        Add pre-hooks

        Args:
            eager_execution: bool: If eager execution is enabled, hooks for
                tracing and executing the action on remote are added.
        """

        # this should be a list as orders matters
        for side_effect in [make_action_side_effect]:
            if side_effect not in self.syft_pre_hooks__[HOOK_ALWAYS]:
                self.syft_pre_hooks__[HOOK_ALWAYS].append(side_effect)

        if eager_execution:
            for side_effect in [send_action_side_effect]:
                if side_effect not in self.syft_pre_hooks__[HOOK_ON_POINTERS]:
                    self.syft_pre_hooks__[HOOK_ON_POINTERS].append(side_effect)

            if trace_action_side_effect not in self.syft_pre_hooks__[HOOK_ALWAYS]:
                self.syft_pre_hooks__[HOOK_ALWAYS].append(trace_action_side_effect)

    def _syft_add_post_hooks__(self, eager_execution: bool) -> None:
        """
        Add post-hooks

        Args:
            eager_execution: bool: If eager execution is enabled, hooks for
                tracing and executing the action on remote are added.
        """
        if eager_execution:
            # this should be a list as orders matters
            for side_effect in [propagate_server_uid]:
                if side_effect not in self.syft_post_hooks__[HOOK_ALWAYS]:
                    self.syft_post_hooks__[HOOK_ALWAYS].append(side_effect)

    def _syft_run_pre_hooks__(
        self, context: PreHookContext, name: str, args: Any, kwargs: Any
    ) -> tuple[PreHookContext, tuple[Any, ...], dict[str, Any]]:
        """Hooks executed before the actual call"""
        result_args, result_kwargs = args, kwargs
        if name in self.syft_pre_hooks__:
            for hook in self.syft_pre_hooks__[name]:
                result = hook(context, *result_args, **result_kwargs)
                if result.is_ok():
                    context, result_args, result_kwargs = result.ok()
                else:
                    logger.debug(f"Pre-hook failed with {result.err()}")
        if name not in self._syft_dont_wrap_attrs():
            if HOOK_ALWAYS in self.syft_pre_hooks__:
                for hook in self.syft_pre_hooks__[HOOK_ALWAYS]:
                    result = hook(context, *result_args, **result_kwargs)
                    if result.is_ok():
                        context, result_args, result_kwargs = result.ok()
                    else:
                        msg = result.err().replace("\\n", "\n")
                        logger.debug(f"Pre-hook failed with {msg}")

        if self.is_pointer:
            if name not in self._syft_dont_wrap_attrs():
                if HOOK_ALWAYS in self.syft_pre_hooks__:
                    for hook in self.syft_pre_hooks__[HOOK_ON_POINTERS]:
                        result = hook(context, *result_args, **result_kwargs)
                        if result.is_ok():
                            context, result_args, result_kwargs = result.ok()
                        else:
                            msg = result.err().replace("\\n", "\n")
                            logger.debug(f"Pre-hook failed with {msg}")

        return context, result_args, result_kwargs

    def _syft_run_post_hooks__(
        self, context: PreHookContext, name: str, result: Any
    ) -> Any:
        """Hooks executed after the actual call"""
        new_result = result
        if name in self.syft_post_hooks__:
            for hook in self.syft_post_hooks__[name]:
                result = hook(context, name, new_result)
                if result.is_ok():
                    new_result = result.ok()
                else:
                    logger.debug(f"Post hook failed with {result.err()}")

        if name not in self._syft_dont_wrap_attrs():
            if HOOK_ALWAYS in self.syft_post_hooks__:
                for hook in self.syft_post_hooks__[HOOK_ALWAYS]:
                    result = hook(context, name, new_result)
                    if result.is_ok():
                        new_result = result.ok()
                    else:
                        logger.debug(f"Post hook failed with {result.err()}")

        if self.is_pointer:
            if name not in self._syft_dont_wrap_attrs():
                if HOOK_ALWAYS in self.syft_post_hooks__:
                    for hook in self.syft_post_hooks__[HOOK_ON_POINTERS]:
                        result = hook(context, name, new_result)
                        if result.is_ok():
                            new_result = result.ok()
                        else:
                            logger.debug(f"Post hook failed with {result.err()}")

        return new_result

    def _syft_output_action_object(
        self, result: Any, context: PreHookContext | None = None
    ) -> Any:
        """Wrap the result in an ActionObject"""
        if issubclass(type(result), ActionObject):
            return result

        constructor = action_type_for_type(result)
        syft_twin_type = TwinMode.NONE
        if context is not None and context.result_twin_type is not None:
            syft_twin_type = context.result_twin_type
        result = constructor(
            syft_twin_type=syft_twin_type,
            syft_action_data_cache=result,
            syft_server_location=self.syft_server_location,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        return result

    def _syft_passthrough_attrs(self) -> list[str]:
        """These attributes are forwarded to the `object` base class."""
        return passthrough_attrs + getattr(self, "syft_passthrough_attrs", [])

    def _syft_dont_wrap_attrs(self) -> list[str]:
        """The results from these attributes are ignored from UID patching."""
        return dont_wrap_output_attrs + getattr(self, "syft_dont_wrap_attrs", [])

    def _syft_get_attr_context(self, name: str) -> Any:
        """Find which instance - Syft ActionObject or the original object - has the requested attribute."""
        defined_on_self = name in self.__dict__ or name in self.__private_attributes__

        # use the custom defined version
        context_self = self
        if not defined_on_self:
            context_self = self.syft_action_data

        return context_self

    def _syft_attr_propagate_ids(
        self, context: PreHookContext, name: str, result: Any
    ) -> Any:
        """Patch the results with the syft_history_hash, server_uid, and result_id."""
        if name in self._syft_dont_wrap_attrs():
            return result

        # Wrap as Syft Object
        result = self._syft_output_action_object(result, context)

        # Propagate History
        if context.action is not None:
            result.syft_history_hash = context.action.syft_history_hash

        # Propagate Syft Server UID
        result.syft_server_uid = context.server_uid

        # Propogate Syft Server Location and Client Verify Key
        result.syft_server_location = context.syft_server_location
        result.syft_client_verify_key = context.syft_client_verify_key

        # Propagate Syft blob storage entry id
        object_attrs = [
            "syft_blob_storage_entry_id",
            "syft_action_data_repr_",
            "syft_action_data_str_",
            "syft_action_data_type",
        ]
        for attr_name in object_attrs:
            attr_value = getattr(context.obj, attr_name, None)
            setattr(result, attr_name, attr_value)

        # Propagate Result ID
        if context.result_id is not None:
            result.id = context.result_id

        return result

    def _syft_wrap_attribute_for_bool_on_nonbools(self, name: str) -> Any:
        """Handle `__getattribute__` for bool casting."""
        if name != "__bool__":
            raise RuntimeError(
                "[_wrap_attribute_for_bool_on_nonbools] Use this only for the __bool__ operator"
            )

        if self.syft_has_bool_attr:
            raise RuntimeError(
                "[_wrap_attribute_for_bool_on_nonbools] self.syft_action_data already implements the bool operator"
            )

        logger.debug("[__getattribute__] Handling bool on nonbools")
        context = PreHookContext(
            obj=self,
            op_name=name,
            syft_server_location=self.syft_server_location,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        context, _, _ = self._syft_run_pre_hooks__(context, name, (), {})

        # no input needs to propagate
        result = self._syft_run_post_hooks__(
            context,
            name,
            any(
                x is not None
                for x in (self.syft_blob_storage_entry_id, self.syft_action_data_cache)
            ),
        )
        result = self._syft_attr_propagate_ids(context, name, result)

        def __wrapper__bool__() -> bool:
            return result

        return __wrapper__bool__

    def _syft_wrap_attribute_for_properties(self, name: str) -> Any:
        """Handle `__getattribute__` for properties."""
        context_self = self._syft_get_attr_context(name)

        if not self.syft_is_property(context_self, name):
            raise RuntimeError(
                "[_wrap_attribute_for_properties] Use this only on properties"
            )
        logger.debug(f"[__getattribute__] Handling property {name}")

        context = PreHookContext(
            obj=self,
            op_name=name,
            action_type=ActionType.GETATTRIBUTE,
            syft_server_location=self.syft_server_location,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        context, _, _ = self._syft_run_pre_hooks__(context, name, (), {})
        # no input needs to propagate
        result = self._syft_run_post_hooks__(
            context, name, self.syft_get_property(context_self, name)
        )

        return self._syft_attr_propagate_ids(context, name, result)

    def _syft_wrap_attribute_for_methods(self, name: str) -> Any:
        """Handle `__getattribute__` for methods."""

        # check for other types that aren't methods, functions etc
        def fake_func(*args: Any, **kwargs: Any) -> Any:
            return ActionDataEmpty(syft_internal_type=self.syft_internal_type)

        logger.debug(f"[__getattribute__] Handling method {name}")
        if (
            issubclass(self.syft_action_data_type, ActionDataEmpty)
            and name not in action_data_empty_must_run
        ):
            original_func = fake_func
        else:
            original_func = getattr(self.syft_action_data, name)

        debug_original_func(name, original_func)

        def _base_wrapper(*args: Any, **kwargs: Any) -> Any:
            context = PreHookContext(
                obj=self,
                op_name=name,
                action_type=ActionType.METHOD,
                syft_server_location=self.syft_server_location,
                syft_client_verify_key=self.syft_client_verify_key,
            )
            context, pre_hook_args, pre_hook_kwargs = self._syft_run_pre_hooks__(
                context, name, args, kwargs
            )

            if has_action_data_empty(args=args, kwargs=kwargs):
                result = fake_func(*args, **kwargs)
            else:
                original_args, original_kwargs = debox_args_and_kwargs(
                    pre_hook_args, pre_hook_kwargs
                )
                result = original_func(*original_args, **original_kwargs)

            post_result = self._syft_run_post_hooks__(context, name, result)
            post_result = self._syft_attr_propagate_ids(context, name, post_result)

            return post_result

        if inspect.ismethod(original_func) or inspect.ismethoddescriptor(original_func):
            logger.debug(f"Running method: {name}")

            def wrapper(_self: Any, *args: Any, **kwargs: Any) -> Any:
                return _base_wrapper(*args, **kwargs)

            wrapper = types.MethodType(wrapper, type(self))
        else:
            logger.debug(f"Running non-method: {name}")

            wrapper = _base_wrapper

        try:
            wrapper.__doc__ = original_func.__doc__
            logger.debug(
                "Found original signature for ",
                name,
                inspect.signature(original_func),
            )
            wrapper.__ipython_inspector_signature_override__ = inspect.signature(
                original_func
            )
        except Exception:
            logger.debug(f"name={name} has no signature")

        # third party
        return wrapper

    def _syft_setattr(self, name: str, value: Any) -> Any:
        args = (name, value)
        kwargs: dict = {}
        op_name = "__setattr__"

        def fake_func(*args: Any, **kwargs: Any) -> Any:
            return ActionDataEmpty(syft_internal_type=self.syft_internal_type)

        if isinstance(
            self.syft_action_data_type, ActionDataEmpty
        ) or has_action_data_empty(args=args, kwargs=kwargs):
            local_func = fake_func
        else:
            local_func = getattr(self.syft_action_data, op_name)

        context = PreHookContext(
            obj=self,
            op_name=op_name,
            action_type=ActionType.SETATTRIBUTE,
            syft_server_location=self.syft_server_location,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        context, pre_hook_args, pre_hook_kwargs = self._syft_run_pre_hooks__(
            context, "__setattr__", args, kwargs
        )

        original_args, _ = debox_args_and_kwargs(pre_hook_args, pre_hook_kwargs)
        val = original_args[1]
        local_func(name, val)
        local_result = self

        post_result = self._syft_run_post_hooks__(context, op_name, local_result)
        post_result = self._syft_attr_propagate_ids(context, op_name, post_result)
        return post_result

    def __getattribute__(self, name: str) -> Any:
        """Called unconditionally to implement attribute accesses for instances of the class.
        If the class also defines __getattr__(), the latter will not be called unless __getattribute__()
        either calls it explicitly or raises an AttributeError.
        This method should return the (computed) attribute value or raise an AttributeError exception.
        In order to avoid infinite recursion in this method, its implementation should always:
         * call the base class method with the same name to access any attributes it needs
            for example : object.__getattribute__(self, name).
         * use the syft/_syft prefix for internal methods.
         * add the method name to the passthrough_attrs.

        Parameters:
            name: str
                The name of the attribute to access.
        """
        # bypass ipython canary verification
        if name == "_ipython_canary_method_should_not_exist_":
            return None

        # bypass certain attrs to prevent recursion issues
        if name.startswith("_syft") or name.startswith("syft"):
            return object.__getattribute__(self, name)

        if name in passthrough_attrs:
            return object.__getattribute__(self, name)

        # third party
        if name in self._syft_passthrough_attrs():
            return object.__getattribute__(self, name)

        # Handle bool operator on nonbools
        if name == "__bool__" and not self.syft_has_bool_attr:
            return self._syft_wrap_attribute_for_bool_on_nonbools(name)

        # check cache first
        if name in methods_to_check_in_cache:
            return getattr(self.syft_action_data_cache, name, None)

        # Handle Properties
        context_self = self._syft_get_attr_context(name)
        if self.syft_is_property(context_self, name):
            return self._syft_wrap_attribute_for_properties(name)

        # Handle anything else
        res = self._syft_wrap_attribute_for_methods(name)
        return res

    @property
    def is_link(self) -> bool:
        return self.syft_action_data_type is ActionDataLink

    def __setattr__(self, name: str, value: Any) -> Any:
        defined_on_self = name in self.__dict__ or name in self.__private_attributes__

        logger.debug(f">> {name} defined_on_self={defined_on_self}")

        # use the custom defined version
        if defined_on_self:
            self.__dict__[name] = value
            return value
        else:
            self._syft_setattr(name, value)
            context_self = self.syft_action_data
            return context_self.__setattr__(name, value)

    # def keys(self) -> KeysView[str]:
    #     return self.syft_action_data.keys()  # type: ignore

    ###### __DUNDER_MIFFLIN__

    # if we do not implement these boiler plate __method__'s then special infix
    # operations like x + y won't trigger __getattribute__
    # unless there is a super special reason we should write no code in these functions
    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        if self.is_mock:
            res = "TwinPointer(Mock)"
        elif self.is_real:
            res = "TwinPointer(Real)"
        elif not self.is_twin:
            res = "Pointer"

        if isinstance(self.syft_action_data_cache, ActionDataEmpty):
            data_repr_ = self.syft_action_data_repr_
        else:
            if inspect.isclass(self.syft_action_data_cache):
                data_repr_ = repr_cls(self.syft_action_data_cache)
            else:
                data_repr_ = (
                    self.syft_action_data_cache._repr_markdown_()
                    if (
                        self.syft_action_data_cache is not None
                        and hasattr(self.syft_action_data_cache, "_repr_markdown_")
                    )
                    else self.syft_action_data_cache.__repr__()
                )

        return f"\n**{res}**\n\n{data_repr_}\n"

    def _data_repr(self) -> str | None:
        if isinstance(self.syft_action_data_cache, ActionDataEmpty):
            data_repr = self.syft_action_data_repr_
        elif inspect.isclass(self.syft_action_data_cache):
            data_repr = repr_cls(self.syft_action_data_cache)
        else:
            data_repr = self.syft_action_data_cache.__repr__()

        return data_repr

    def __repr__(self) -> str:
        if self.is_mock:
            res = "TwinPointer(Mock)"
        elif self.is_real:
            res = "TwinPointer(Real)"
        if not self.is_twin:
            res = "Pointer"
        data_repr = self._data_repr()
        return f"{res}:\n{data_repr}"

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.__call__(*args, **kwds)

    def __str__(self) -> str:
        if not inspect.isclass(self):
            return self.__str__()
        else:
            return self.syft_action_data_str_

    def __len__(self) -> int:
        return self.__len__()

    def __hash__(self, *args: Any, **kwargs: Any) -> int:
        return super().__hash__(*args, **kwargs)

    def __getitem__(self, key: Any) -> Any:
        return self._syft_output_action_object(self.__getitem__(key))

    def __setitem__(self, key: Any, value: Any) -> None:
        return self.__setitem__(key, value)

    def __contains__(self, key: Any) -> bool:
        return self.__contains__(key)

    def __bool__(self) -> bool:
        return self.__bool__()

    def __add__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__add__(other))

    def __sub__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__sub__(other))

    def __mul__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__mul__(other))

    def __matmul__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__matmul__(other))

    def __eq__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__eq__(other))

    def __ne__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__ne__(other))

    def __lt__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__lt__(other))

    def __gt__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__gt__(other))

    def __le__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__le__(other))

    def __ge__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__ge__(other))

    def __delattr__(self, key: Any) -> None:
        self.__delattr__(key)

    def __delitem__(self, key: Any) -> None:
        self.__delitem__(key)

    def __invert__(self) -> Any:
        return self._syft_output_action_object(self.__invert__())

    def __round__(self) -> Any:
        return self._syft_output_action_object(self.__round__())

    def __pos__(self) -> Any:
        return self._syft_output_action_object(self.__pos__())

    def __trunc__(self) -> Any:
        return self._syft_output_action_object(self.__trunc__())

    def __divmod__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__divmod__(other))

    def __floordiv__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__floordiv__(other))

    def __mod__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__mod__(other))

    def __abs__(self) -> Any:
        return self._syft_output_action_object(self.__abs__())

    def __neg__(self) -> Any:
        return self._syft_output_action_object(self.__neg__())

    def __or__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__or__(other))

    def __and__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__and__(other))

    def __xor__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__xor__(other))

    def __pow__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__pow__(other))

    def __truediv__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__truediv__(other))

    def __lshift__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__lshift__(other))

    def __rshift__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rshift__(other))

    def __iter__(self) -> Any:
        return self._syft_output_action_object(self.__iter__())

    def __next__(self) -> Any:
        return self._syft_output_action_object(self.__next__())

    # r ops
    # we want the underlying implementation so we should just call into __getattribute__
    def __radd__(self, other: Any) -> Any:
        return self.__radd__(other)

    def __rsub__(self, other: Any) -> Any:
        return self.__rsub__(other)

    def __rmul__(self, other: Any) -> Any:
        return self.__rmul__(other)

    def __rmatmul__(self, other: Any) -> Any:
        return self.__rmatmul__(other)

    def __rmod__(self, other: Any) -> Any:
        return self.__rmod__(other)

    def __ror__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__ror__(other))

    def __rand__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rand__(other))

    def __rxor__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rxor__(other))

    def __rpow__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rpow__(other))

    def __rtruediv__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rtruediv__(other))

    def __rfloordiv__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rfloordiv__(other))

    def __rlshift__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rlshift__(other))

    def __rrshift__(self, other: Any) -> Any:
        return self._syft_output_action_object(self.__rrshift__(other))


@serializable()
class AnyActionObject(ActionObject):
    """
    This is a catch-all class for all objects that are not
    defined in the `action_types` dictionary.
    """

    __canonical_name__ = "AnyActionObject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: ClassVar[type[Any]] = NoneType  # type: ignore
    # syft_passthrough_attrs: List[str] = []
    syft_dont_wrap_attrs: list[str] = ["__str__", "__repr__", "syft_action_data_str_"]
    syft_action_data_str_: str = ""

    def __float__(self) -> float:
        return float(self.syft_action_data)

    def __int__(self) -> float:
        return int(self.syft_action_data)


action_types[Any] = AnyActionObject


def debug_original_func(name: str, func: Callable) -> None:
    logger.debug(f"{name} func is:")
    logger.debug(f"inspect.isdatadescriptor = {inspect.isdatadescriptor(func)}")
    logger.debug(f"inspect.isgetsetdescriptor = {inspect.isgetsetdescriptor(func)}")
    logger.debug(f"inspect.isfunction = {inspect.isfunction(func)}")
    logger.debug(f"inspect.isbuiltin = {inspect.isbuiltin(func)}")
    logger.debug(f"inspect.ismethod = {inspect.ismethod(func)}")
    logger.debug(f"inspect.ismethoddescriptor = {inspect.ismethoddescriptor(func)}")


def is_action_data_empty(obj: Any) -> bool:
    return isinstance(obj, AnyActionObject) and issubclass(
        obj.syft_action_data_type, ActionDataEmpty
    )


def has_action_data_empty(args: Any, kwargs: Any) -> bool:
    for a in args:
        if is_action_data_empty(a):
            return True

    for a in kwargs.values():
        if is_action_data_empty(a):
            return True
    return False



# stdlib

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...types.uid import UID
from ...util.telemetry import instrument
from .dataset import Dataset
from .dataset import DatasetUpdate

NamePartitionKey = PartitionKey(key="name", type_=str)
ActionIDsPartitionKey = PartitionKey(key="action_ids", type_=list[UID])


@instrument
@serializable(canonical_name="DatasetStash", version=1)
class DatasetStash(BaseUIDStoreStash):
    object_type = Dataset
    settings: PartitionSettings = PartitionSettings(
        name=Dataset.__canonical_name__, object_type=Dataset
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_by_name(
        self, credentials: SyftVerifyKey, name: str
    ) -> Result[Dataset | None, str]:
        qks = QueryKeys(qks=[NamePartitionKey.with_obj(name)])
        return self.query_one(credentials=credentials, qks=qks)

    def update(
        self,
        credentials: SyftVerifyKey,
        dataset_update: DatasetUpdate,
        has_permission: bool = False,
    ) -> Result[Dataset, str]:
        res = self.check_type(dataset_update, DatasetUpdate)
        # we dont use and_then logic here as it is hard because of the order of the arguments
        if res.is_err():
            return res
        return super().update(credentials=credentials, obj=res.ok())

    def search_action_ids(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[list[Dataset], str]:
        qks = QueryKeys(qks=[ActionIDsPartitionKey.with_obj(uid)])
        return self.query_all(credentials=credentials, qks=qks)

    def get_all(
        self,
        credentials: SyftVerifyKey,
        order_by: PartitionKey | None = None,
        has_permission: bool = False,
    ) -> Ok[list] | Err[str]:
        result = super().get_all(credentials, order_by, has_permission)

        if result.is_err():
            return result
        filtered_datasets = [
            dataset for dataset in result.ok_value if not dataset.to_be_deleted
        ]
        return Ok(filtered_datasets)



# stdlib

# third party
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...util.telemetry import instrument
from .user_code import CodeHashPartitionKey
from .user_code import ServiceFuncNamePartitionKey
from .user_code import SubmitTimePartitionKey
from .user_code import UserCode
from .user_code import UserVerifyKeyPartitionKey


@instrument
@serializable(canonical_name="UserCodeStash", version=1)
class UserCodeStash(BaseUIDStoreStash):
    object_type = UserCode
    settings: PartitionSettings = PartitionSettings(
        name=UserCode.__canonical_name__, object_type=UserCode
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_all_by_user_verify_key(
        self, credentials: SyftVerifyKey, user_verify_key: SyftVerifyKey
    ) -> Result[list[UserCode], str]:
        qks = QueryKeys(qks=[UserVerifyKeyPartitionKey.with_obj(user_verify_key)])
        return self.query_one(credentials=credentials, qks=qks)

    def get_by_code_hash(
        self, credentials: SyftVerifyKey, code_hash: str
    ) -> Result[UserCode | None, str]:
        qks = QueryKeys(qks=[CodeHashPartitionKey.with_obj(code_hash)])
        return self.query_one(credentials=credentials, qks=qks)

    def get_by_service_func_name(
        self, credentials: SyftVerifyKey, service_func_name: str
    ) -> Result[list[UserCode], str]:
        qks = QueryKeys(qks=[ServiceFuncNamePartitionKey.with_obj(service_func_name)])
        return self.query_all(
            credentials=credentials, qks=qks, order_by=SubmitTimePartitionKey
        )
